from autocti.data import mask as msk

import numpy as np

class CIMask(msk.Mask):

    def __new__(cls, array, *args, **kwargs):
        mask = np.array(array, dtype='bool').view(cls)
        return mask

    @classmethod
    def masked_parallel_front_edge_from_ci_frame(cls, shape, ci_frame, rows, invert=False):

        front_edge_regions = ci_frame.parallel_front_edge_regions_from_frame(rows=rows)
        mask = np.full(shape, False)

        for region in front_edge_regions:
            mask[region.y0:region.y1, region.x0:region.x1] = True

        if invert: mask = np.invert(mask)

        return CIMask(array=mask.astype('bool'), frame_geometry=ci_frame.frame_geometry)

    @classmethod
    def masked_parallel_trails_from_ci_frame(cls, shape, ci_frame, rows, invert=False):

        trails_regions = ci_frame.parallel_trails_regions_from_frame(rows=rows)
        mask = np.full(shape, False)

        for region in trails_regions:
            mask[region.y0:region.y1, region.x0:region.x1] = True

        if invert: mask = np.invert(mask)

        return CIMask(array=mask.astype('bool'), frame_geometry=ci_frame.frame_geometry)

    @classmethod
    def masked_serial_front_edge_from_ci_frame(cls, shape, ci_frame, columns, invert=False):

        front_edge_regions = ci_frame.serial_front_edge_regions_from_frame(columns=columns)
        mask = np.full(shape, False)

        for region in front_edge_regions:
            mask[region.y0:region.y1, region.x0:region.x1] = True

        if invert: mask = np.invert(mask)

        return CIMask(array=mask.astype('bool'), frame_geometry=ci_frame.frame_geometry)

    @classmethod
    def masked_serial_trails_from_ci_frame(cls, shape, ci_frame, columns, invert=False):

        trails_regions = ci_frame.serial_trails_regions_from_frame(columns=columns)
        mask = np.full(shape, False)

        for region in trails_regions:
            mask[region.y0:region.y1, region.x0:region.x1] = True

        if invert: mask = np.invert(mask)

        return CIMask(array=mask.astype('bool'), frame_geometry=ci_frame.frame_geometry)


#
# Copyright (C) 2012-2020 Euclid Science Ground Segment
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3.0 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
File: python/SHE_ArCTIC/ImageIO.py

Created on: 04/23/18
Author: James Nightingale
"""

import os

import numpy as np
from astropy.io import fits


def make_path_if_does_not_exist(path):
    if os.path.exists(path) is False:
        os.makedirs(path)


def numpy_array_2d_to_fits(array_2d, file_path, overwrite=False):
    """Write a 2D NumPy array to a .fits file.

    Parameters
    ----------
    array_2d : ndarray
        The 2D array that is written to fits.
    file_path : str
        The full path of the file that is output, including the file name and '.fits' extension.
    overwrite : bool
        If True and a file already exists with the input file_path the .fits file is overwritten. If False, an error \
        will be raised.

    Returns
    -------
    None

    Examples
    --------
    array_2d = np.ones((5,5))
    numpy_array_to_fits(array=array_2d, file_path='/path/to/file/filename.fits', overwrite=True)
    """
    if overwrite and os.path.exists(file_path):
        os.remove(file_path)

    new_hdr = fits.Header()
    hdu = fits.PrimaryHDU(array_2d, new_hdr)
    hdu.writeto(file_path)


def numpy_array_2d_from_fits(file_path, hdu):
    """Read a 2D NumPy array to a .fits file.

    After loading the NumPy array, the array is flipped upside-down using np.flipud. This is so that the arrays \
    appear the same orientation as .fits files loaded in DS9.

    Parameters
    ----------
    file_path : str
        The full path of the file that is loaded, including the file name and '.fits' extension.
    hdu : int
        The HDU extension of the array that is loaded from the .fits file.

    Returns
    -------
    ndarray
        The NumPy array that is loaded from the .fits file.

    Examples
    --------
    array_2d = numpy_array_from_fits(file_path='/path/to/file/filename.fits', hdu=0)
    """
    hdu_list = fits.open(file_path)
    return np.array(hdu_list[hdu].data)


"""
File: python/VIS_CTICalibrate/ChargeInjectPattern.py

Created on: 02/14/18
Author: James Nightingale
"""

import numpy as np

from autocti import exc
from autocti.charge_injection import ci_frame


def uniform_from_lists(normalizations, regions):
    """Setup the collection of patterns from lists of uniform ci_pattern properties

    Params
    -----------
    normalizations : list
        The normalization in each charge injection ci_pattern.
    regions : [(int, int, int, int)]
        The regions each charge injection ci_pattern appears. This is identical across all images.
    """
    return list(map(lambda n: CIPatternUniform(n, regions), normalizations))


def non_uniform_from_lists(normalizations, regions, row_slopes):
    """Setup the collection of patterns from lists of non-uniform ci_pattern properties

    Params
    -----------
    normalizations : [float]
        The normalization in each charge injection ci_pattern.
    regions : [(int, int, int, int)]
        The regions each charge injection ci_pattern appears. This is identical across all images.
    row_slopes : [float]
        The power-law slopes of non-uniformity in the rows of each charge injection profile.
    """
    return list(map(lambda n, s: CIPatternNonUniform(n, regions, s), normalizations, row_slopes))


class CIPattern(object):

    def __init__(self, normalization, regions):
        """ Abstract base class for a charge injection ci_pattern, which defines the regions charge injections appears \
         on a charge-injection ci_frame, the input normalization and other properties.

        Parameters
        -----------
        normalization : float
            The normalization of the charge injection lines.
        regions: [(int,)]
            A list of the integer coordinates specifying the corners of each charge injection region \
            (top-row, bottom-row, left-column, right-column).
        """
        self.normalization = normalization
        self.regions = list(map(ci_frame.Region, regions))

    def check_pattern_is_within_image_dimensions(self, dimensions):

        for region in self.regions:

            if region.y1 > dimensions[0] or region.x1 > dimensions[1]:
                raise exc.CIPatternException(
                    'The charge injection ci_pattern regions are bigger than the image image_shape')

class CIPatternUniform(CIPattern):
    """ A uniform charge injection ci_pattern, which is defined by the regions it appears on the charge injection \
        ci_frame and its normalization.

    """

    def ci_pre_cti_from_shape(self, shape):
        """Compute the pre-cti image of the uniform charge injection ci_pattern.

        This is performed by going to each charge injection region and adding the charge injection normalization value.

        Parameters
        -----------
        shape : (int,)
            The image_shape of the ci_pre_ctis to be created.
        """

        self.check_pattern_is_within_image_dimensions(shape)

        ci_pre_cti = np.zeros(shape)

        for region in self.regions:
            ci_pre_cti[region.slice] += self.normalization

        return ci_pre_cti

    """ Class to simulate a charge injection image with a uniform charge injection ci_pattern."""

    def simulate_ci_pre_cti(self, shape):
        """Use this charge injection ci_pattern to generate a pre-cti charge injection image. This is performed by \
        going to its charge injection regions and adding the charge injection normalization value.

        Parameters
        -----------
        shape : (int, int)
            The image_shape of the ci_pre_ctis to be created.
        """

        self.check_pattern_is_within_image_dimensions(shape)

        ci_pre_cti = np.zeros(shape)

        for region in self.regions:
            ci_pre_cti[region.slice] += self.normalization

        return ci_pre_cti


class CIPatternNonUniform(CIPattern):

    def __init__(self, normalization, regions, row_slope):
        """ A non-uniform charge injection ci_pattern, which is defined by the regions it appears on a charge injection
        ci_frame and its average normalization.

        Non-uniformity across the columns of a charge injection ci_pattern is due to spikes / drops in the current that
        injects the charge. This is a noisy process, leading to non-uniformity with no regularity / smoothness. Thus,
        it cannot be modeled with an analytic profile, and must be assumed as prior-knowledge about the charge
        injection electronics or estimated from the observed charge injection ci_data.

        Non-uniformity across the rows of a charge injection ci_pattern is due to a drop-off in voltage in the current.
        Therefore, it appears smooth and be modeled as an analytic function, which this code assumes is a
        power-law with slope row_slope.

        Parameters
        -----------
        normalization : float
            The normalization of the charge injection region.
        regions : [(int,)]
            A list of the integer coordinates specifying the corners of each charge injection region
            (top-row, bottom-row, left-column, right-column).
        row_slope : float
            The power-law slope of non-uniformity in the row charge injection profile.
        """
        super(CIPatternNonUniform, self).__init__(normalization, regions)
        self.row_slope = row_slope

    def ci_pre_cti_from_ci_image_and_mask(self, ci_image, mask):
        """Compute the pre-cti image of this non-uniform charge injection ci_pattern.

        This is performed by estimating the charge injection of each column, by taking the average value of all charge \
        regions in that column (including non-uniformity across the rows). The mask is used to remove cosmic rays from \
        this averaging.

        Parameters
        -----------
        ci_image : ndarray
            2D array of ci_pre_ctis ci_data the column non-uniformity is estimated from.
        mask : ndarray
            2D array of masked ci_pre_ctis pixels, used to mask cosmic rays.
        """

        dimensions = ci_image.shape

        self.check_pattern_is_within_image_dimensions(dimensions)

        ci_pre_cti = np.zeros(dimensions)

        for column_number in range(dimensions[1]):

            means_of_columns = self.mean_charge_in_all_image_columns(column=ci_image[:, column_number],
                                                                     column_mask=mask[:, column_number])

            filtered = list(filter(None, means_of_columns))
            if len(filtered) == 0:
                raise exc.CIPatternException(
                    'All Pixels in a charge injection region were flagged as masked - code does'
                    'not currently handle such a circumstance')

            overall_mean = np.mean(filtered)

            for region in self.regions:
                ci_pre_cti[region.y_slice, column_number] = overall_mean

        return ci_pre_cti

    def mean_charge_in_all_image_columns(self, column, column_mask):
        """For one column of a charge injection image, measure the mean charge in each column of each charge injection \
        region.

        Parameters
        -----------
        column : ndarray
            1D array of the ci_pre_ctis column.
        column_mask : ndarray
            1D array of the column's masked pixels, to mask cosmic rays.
         """

        means_of_columns = []

        for region in self.regions:
            means_of_columns.append(self.mean_charge_in_column(column[region.y_slice].flatten(),
                                                               column_mask[region.y_slice].flatten()))

        return means_of_columns

    def mean_charge_in_column(self, column, column_mask):
        """Measure the mean charge in a column of a non-uniform charge injection region (accounting for row \
        non-uniformity if present). Cosmic rays can be masked.

        Parameters
        -----------
        column : ndarray
            1D array of the column of the charge injection region.
        column_mask : ndarray
            1D array of the column's masked pixels, to mask cosmic rays.
        """
        column_size = len(column)

        if sum(column_mask) == column_size:
            return None

        model_column = self.generate_column(column_size, self.normalization)
        return np.mean(np.ma.masked_array(column - model_column + self.normalization, column_mask))

    def generate_column(self, size, normalization):
        """Generate a column of non-uniform charge, including row non-uniformity.

        The pixel-numbering used to generate non-uniformity across the charge injection rows runs from 1 -> size

        Parameters
        -----------
        size : int
            The size of the non-uniform column of charge
        normalization : float
            The input normalization of the column's charge e.g. the level of charge injected.

        """
        return normalization * (np.arange(1, size + 1)) ** self.row_slope

    def simulate_ci_pre_cti(self, shape, ci_seed=-1, column_deviation=0.0,
                            maximum_normalization=np.inf):
        """Use this charge injection ci_pattern to generate a pre-cti charge injection image. This is performed by going \
        to its charge injection regions and adding its non-uniform charge distribution.

        For one column of a non-uniform charge injection ci_pre_ctis, it is assumed that each non-uniform charge \
        injection region has the same overall normalization value (after drawing this value randomly from a Gaussian \
        distribution). Physically, this is true provided the spikes / troughs in the current that cause \
        non-uniformity occur in an identical fashion for the generation of each charge injection region.

        Parameters
        -----------
        column_deviation
        shape
            The image_shape of the ci_pre_ctis to be created.
        maximum_normalization

        ci_seed : int
            Input ci_seed for the random number generator to give reproducible results. A new ci_seed is always used for each \
            ci_pre_ctis, ensuring each non-uniform ci_region has the same column non-uniformity ci_pattern.
        """

        self.check_pattern_is_within_image_dimensions(shape)

        ci_pre_cti = np.zeros(shape)

        if ci_seed == -1:
            ci_seed = np.random.randint(0, int(1e9))  # Use one ci_seed, so all regions have identical column
            # non-uniformity.

        for region in self.regions:
            ci_pre_cti[region.slice] += \
                self.simulate_region(region_dimensions=region.shape, ci_seed=ci_seed, column_deviation=column_deviation,
                                     maximum_normalization=maximum_normalization)

        return ci_pre_cti

    def simulate_region(self, region_dimensions, ci_seed, column_deviation=0.0, maximum_normalization=np.inf):
        """Generate the non-uniform charge distribution of a charge injection region. This includes non-uniformity \
        across both the rows and columns of the charge injection region.

        Before adding non-uniformity to the rows and columns, we assume an input charge injection level \
        (e.g. the average current being injected). We then simulate non-uniformity in this region.

        Non-uniformity in the columns is caused by sharp peaks and troughs in the input charge current. To simulate  \
        this, we change the normalization of each column by drawing its normalization value from a Gaussian \
        distribution which has a mean of the input normalization and standard deviation *column_deviation*. The seed \
        of the random number generator ensures that the non-uniform charge injection update_via_regions of each ci_pre_ctis \
        are identical.

        Non-uniformity in the rows is caused by the charge smoothly decreasing as the injection is switched off. To \
        simulate this, we assume the charge level as a function of row number is not flat but defined by a \
        power-law with slope *row_slope*.

        Non-uniform charge injection images are generated using the function *simulate_pre_cti*, which uses this \
        function.

        Parameters
        -----------
        maximum_normalization
        column_deviation
        region_dimensions : (int, int)
            The size of the non-uniform charge injection region.
        ci_seed : int
            Input seed for the random number generator to give reproducible results.
        """

        np.random.seed(ci_seed)

        ci_rows = region_dimensions[0]
        ci_columns = region_dimensions[1]
        ci_region = np.zeros(region_dimensions)

        for column_number in range(ci_columns):

            column_normalization = 0
            while column_normalization <= 0 or column_normalization >= maximum_normalization:
                column_normalization = np.random.normal(self.normalization, column_deviation)

            ci_region[0:ci_rows, column_number] = self.generate_column(size=ci_rows, normalization=column_normalization)

        return ci_region



#
# Copyright (C) 2012-2020 Euclid Science Ground Segment
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3.0 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
File: tests/python/ArcticParams_test.py

Created on: 02/13/18
Author: James Nightingale
"""

import os
import shutil

import pytest

from autocti.model import arctic_params


class TestParams:
    
    def test__1_species__sets_values_correctly(self):
        
        parallel_1_species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)

        parameters = arctic_params.ArcticParams(parallel_species=parallel_1_species)

        assert type(parameters) == arctic_params.ArcticParams
        assert type(parameters.parallel_species) == arctic_params.Species

        assert parameters.parallel_species.trap_density == 0.1
        assert parameters.parallel_species.trap_lifetime == 1.0
    
    def test__parallel_and_serial_species__sets_value_correctly(self):
        
        parallel_species_0 = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
        parallel_species_1 = arctic_params.Species(trap_density=0.2, trap_lifetime=2.0)

        parallel_ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=0.2,
                                         well_fill_beta=0.8, well_fill_gamma=2.0)

        serial_species_0 = arctic_params.Species(trap_density=0.3, trap_lifetime=3.0)
        serial_species_1 = arctic_params.Species(trap_density=0.4, trap_lifetime=4.0)

        serial_ccd = arctic_params.CCD(well_notch_depth=1.02, well_fill_alpha=1.1,
                                       well_fill_beta=1.4, well_fill_gamma=1.6)

        parameters = arctic_params.ArcticParams(parallel_species=[parallel_species_0, parallel_species_1],
                                                serial_species=[serial_species_0, serial_species_1], 
                                                parallel_ccd=parallel_ccd, serial_ccd=serial_ccd)

        assert type(parameters) == arctic_params.ArcticParams
        assert type(parameters.parallel_species[0]) == arctic_params.Species
        assert type(parameters.parallel_species[1]) == arctic_params.Species
        assert type(parameters.parallel_ccd) == arctic_params.CCD

        assert parameters.parallel_species[0].trap_density == 0.1
        assert parameters.parallel_species[0].trap_lifetime == 1.0
        assert parameters.parallel_species[1].trap_density == 0.2
        assert parameters.parallel_species[1].trap_lifetime == 2.0
        assert parameters.parallel_ccd.well_notch_depth == 0.01
        assert parameters.parallel_ccd.well_fill_alpha == 0.2
        assert parameters.parallel_ccd.well_fill_beta == 0.8
        assert parameters.parallel_ccd.well_fill_gamma == 2.0
        
        assert type(parameters) == arctic_params.ArcticParams
        assert type(parameters.serial_species[0]) == arctic_params.Species
        assert type(parameters.serial_species[1]) == arctic_params.Species
        assert type(parameters.serial_ccd) == arctic_params.CCD

        assert parameters.serial_species[0].trap_density == 0.3
        assert parameters.serial_species[0].trap_lifetime == 3.0
        assert parameters.serial_species[1].trap_density == 0.4
        assert parameters.serial_species[1].trap_lifetime == 4.0
        assert parameters.serial_ccd.well_notch_depth == 1.02
        assert parameters.serial_ccd.well_fill_alpha == 1.1
        assert parameters.serial_ccd.well_fill_beta == 1.4
        assert parameters.serial_ccd.well_fill_gamma == 1.6


class TestParallelDensityVary:

    def test_1_species__density_01__1000_column_pixels__1_row_pixel_so_100_traps__posison_density_near_01(self):
        parallel_vary = arctic_params.Species.poisson_species(
            species=list(map(lambda density: arctic_params.Species(trap_density=density, trap_lifetime=1.), (0.1,))),
            shape=(1000, 1),
            seed=1)

        assert [species.trap_density for species in parallel_vary] == [0.098]

    def test__1_species__density_1__1000_column_pixels_so_1000_traps__1_row_pixel__posison_value_is_near_1(self):
        parallel_vary = arctic_params.Species.poisson_species(
            species=list(map(lambda density: arctic_params.Species(trap_density=density, trap_lifetime=1.), (1.0,))),
            shape=(1000, 1),
            seed=1)

        assert [species.trap_density for species in parallel_vary] == [0.992]

    def test__1_species__density_1___2_row_pixels__posison_value_is_near_1(self):
        parallel_vary = arctic_params.Species.poisson_species(
            species=list(map(lambda density: arctic_params.Species(trap_density=density, trap_lifetime=1.), (1.0,))),
            shape=(1000, 2),
            seed=1)

        assert [species.trap_density for species in parallel_vary] == [0.992, 0.962]

    def test__2_species__1_row_pixel__poisson_for_each_species_drawn(self):
        parallel_vary = arctic_params.Species.poisson_species(species=list(
            map(lambda density: arctic_params.Species(trap_density=density, trap_lifetime=1.), (1.0, 2.0))),
            shape=(1000, 1),
            seed=1)

        assert [species.trap_density for species in parallel_vary] == [0.992, 1.946]

    def test__2_species__2_row_pixel__poisson_for_each_species_drawn(self):
        parallel_vary = arctic_params.Species.poisson_species(species=list(
            map(lambda density: arctic_params.Species(trap_density=density, trap_lifetime=1.), (1.0, 2.0))),
            shape=(1000, 2),
            seed=1)

        assert [species.trap_density for species in parallel_vary] == [0.992, 1.946, 0.968, 1.987]

    def test__same_as_above_but_3_species_and_new_values(self):
        parallel_vary = arctic_params.Species.poisson_species(species=list(
            map(lambda density: arctic_params.Species(trap_density=density, trap_lifetime=1.), (1.0, 2.0, 0.1))),
            shape=(1000, 3),
            seed=1)

        assert [species.trap_density for species in parallel_vary] == [0.992, 1.946, 0.09, 0.991, 1.99, 0.098, 0.961,
                                                                       1.975, 0.113]



from autocti.charge_injection import ci_frame

def parallel_shape_and_frame_geometry_from_ci_data_resolution(ci_data_resolution):

    if ci_data_resolution is 'high_resolution':

        shape = (2316, 2119)
        frame_geometry = ci_frame.FrameGeometry(corner=(0.0, 0.0),
                                                parallel_overscan=ci_frame.Region((2296, 2316, 51, 2099)),
                                                serial_prescan=ci_frame.Region((0, 2316, 0, 51)),
                                                serial_overscan=ci_frame.Region((0, 2296, 2099, 2119)))

        return shape, frame_geometry

    elif ci_data_resolution is 'mid_resolution':

        shape = (2316, 1034)
        frame_geometry = ci_frame.FrameGeometry(corner=(0.0, 0.0),
                                                parallel_overscan=ci_frame.Region((2296, 2316, 51, 1014)),
                                                serial_prescan=ci_frame.Region((0, 2316, 0, 51)),
                                                serial_overscan=ci_frame.Region((0, 2296, 1014, 1034)))

        return shape, frame_geometry

    elif ci_data_resolution is 'low_resolution':

        shape = (2316, 517)
        frame_geometry = ci_frame.FrameGeometry(corner=(0.0, 0.0),
                                                parallel_overscan=ci_frame.Region((2296, 2316, 51, 497)),
                                                serial_prescan=ci_frame.Region((0, 2316, 0, 51)),
                                                serial_overscan=ci_frame.Region((0, 2296, 497, 517)))

        return shape, frame_geometry

def serial_shape_and_frame_geometry_from_ci_data_resolution(ci_data_resolution):

    if ci_data_resolution is 'high_resolution':

        shape = (2066, 2119)
        frame_geometry = ci_frame.FrameGeometry(corner=(0.0, 0.0),
                                                parallel_overscan=ci_frame.Region((2296, 2316, 51, 2099)),
                                                serial_prescan=ci_frame.Region((0, 2316, 0, 51)),
                                                serial_overscan=ci_frame.Region((0, 2296, 2099, 2119)))

        return shape, frame_geometry

    elif ci_data_resolution is 'mid_resolution':

        shape = (1033, 2119)
        frame_geometry = ci_frame.FrameGeometry(corner=(0.0, 0.0),
                                                parallel_overscan=ci_frame.Region((1138, 1158, 51, 2099)),
                                                serial_prescan=ci_frame.Region((0, 1158, 0, 51)),
                                                serial_overscan=ci_frame.Region((0, 1138, 2099, 2119)))

        return shape, frame_geometry

    elif ci_data_resolution is 'low_resolution':

        shape = (517, 2119)
        frame_geometry = ci_frame.FrameGeometry(corner=(0.0, 0.0),
                                                parallel_overscan=ci_frame.Region((559, 579, 51, 2099)),
                                                serial_prescan=ci_frame.Region((0, 579, 0, 51)),
                                                serial_overscan=ci_frame.Region((0, 559, 2099, 2119)))

        return shape, frame_geometry

def parallel_and_serial_shape_and_frame_geometry():

    shape = (2316, 2119)
    frame_geometry = ci_frame.FrameGeometry(corner=(0.0, 0.0),
                                            parallel_overscan=ci_frame.Region((2296, 2316, 51, 2099)),
                                            serial_prescan=ci_frame.Region((0, 2316, 0, 51)),
                                            serial_overscan=ci_frame.Region((0, 2296, 2099, 2119)))

    return shape, frame_geometry


from autofit.core import model_mapper as mm
from autocti.model import arctic_params as ap
import workspace_jam.scripts.requirements as req
import workspace_jam.scripts.requirements.requirement_converter as convert

import matplotlib.pyplot as plt
import numpy as np

model_mapper = mm.ModelMapper(serial=ap.SerialThreeSpecies)
model_mapper.serial.well_fill_alpha = 0.0
model_mapper.serial.well_fill_gamma = 0.0

input_model = [0.01, 0.03, 0.9, 0.8, 3.5, 20.0, 1.0e-4, 0.58]
input_rho = np.array([0.01, 0.03, 0.9])
input_tau = np.array([0.8, 3.5, 20.0])

sigma_limit = 2.0

delta_ellipticity_true = convert.convert_to_ellipticity(rho=input_rho, tau=input_tau)

print('Model True')
print(delta_ellipticity_true)
print(np.sum(delta_ellipticity_true))
print()

most_likely_lr, most_probable_lr, lower_limits_lr, upper_limits_lr, lower_error_lr, upper_error_lr, error_pecision_lr = \
    req.setup_multinest_results(model_mapper=model_mapper, input_model=input_model, sigma_limit=sigma_limit,
                                name='Results/Serial_Tight_lr/')

model_rho_lr = np.array([most_probable_lr[0], most_probable_lr[1]])
model_tau_lr = np.array([most_probable_lr[2], most_probable_lr[3]])
delta_ellipticity_model_lr = convert.convert_to_ellipticity(rho=model_rho_lr, tau=model_tau_lr)

upper_rho_lr = np.array([upper_limits_lr[0], upper_limits_lr[1]])
upper_tau_lr = np.array([upper_limits_lr[2], upper_limits_lr[3]])
delta_ellipticity_upper_lr = convert.convert_to_ellipticity(rho=upper_rho_lr, tau=upper_tau_lr)

lower_rho_lr = np.array([lower_limits_lr[0], lower_limits_lr[1]])
lower_tau_lr = np.array([lower_limits_lr[2], lower_limits_lr[3]])
delta_ellipticity_lower_lr = convert.convert_to_ellipticity(rho=lower_rho_lr, tau=lower_tau_lr)

print('Induced Delta Elliptcitiy / Requirement')
print((np.sum(delta_ellipticity_model_lr) - np.sum(delta_ellipticity_true)) / (1.1e-4))
print((np.sum(delta_ellipticity_upper_lr) - np.sum(delta_ellipticity_true)) / (1.1e-4))
print((np.sum(delta_ellipticity_lower_lr) - np.sum(delta_ellipticity_true)) / (1.1e-4))
print()

ellipticity_requirement_lr = (np.sum(delta_ellipticity_model_lr) - np.sum(delta_ellipticity_true)) / (1.1e-4)
ellipticity_requirement_upper_lr = (np.sum(delta_ellipticity_upper_lr) - np.sum(delta_ellipticity_true)) / (1.1e-4)
ellipticity_requirement_lower_lr = (np.sum(delta_ellipticity_lower_lr) - np.sum(delta_ellipticity_true)) / (1.1e-4)
ellipticity_error_upper_lr = ellipticity_requirement_upper_lr - ellipticity_requirement_lr
ellipticity_error_lower_lr = ellipticity_requirement_lr - ellipticity_requirement_lower_lr

most_likely_mr, most_probable_mr, lower_limits_mr, upper_limits_mr, lower_error_mr, upper_error_mr, error_pecision_mr = \
    req.setup_multinest_results(model_mapper=model_mapper, input_model=input_model, sigma_limit=sigma_limit,
                                name='Results/Serial_Tight_mr/')

model_rho_mr = np.array([most_probable_mr[0], most_probable_mr[1]])
model_tau_mr = np.array([most_probable_mr[2], most_probable_mr[3]])
delta_ellipticity_model_mr = convert.convert_to_ellipticity(rho=model_rho_mr, tau=model_tau_mr)

upper_rho_mr = np.array([upper_limits_mr[0], upper_limits_mr[1]])
upper_tau_mr = np.array([upper_limits_mr[2], upper_limits_mr[3]])
delta_ellipticity_upper_mr = convert.convert_to_ellipticity(rho=upper_rho_mr, tau=upper_tau_mr)

lower_rho_mr = np.array([lower_limits_mr[0], lower_limits_mr[1]])
lower_tau_mr = np.array([lower_limits_mr[2], lower_limits_mr[3]])
delta_ellipticity_lower_mr = convert.convert_to_ellipticity(rho=lower_rho_mr, tau=lower_tau_mr)

print('Induced Delta Elliptcitiy / Requirement')
print((np.sum(delta_ellipticity_model_mr) - np.sum(delta_ellipticity_true)) / (1.1e-4))
print((np.sum(delta_ellipticity_upper_mr) - np.sum(delta_ellipticity_true)) / (1.1e-4))
print((np.sum(delta_ellipticity_lower_mr) - np.sum(delta_ellipticity_true)) / (1.1e-4))
print()

ellipticity_requirement_mr = (np.sum(delta_ellipticity_model_mr) - np.sum(delta_ellipticity_true)) / (1.1e-4)
ellipticity_requirement_upper_mr = (np.sum(delta_ellipticity_upper_mr) - np.sum(delta_ellipticity_true)) / (1.1e-4)
ellipticity_requirement_lower_mr = (np.sum(delta_ellipticity_lower_mr) - np.sum(delta_ellipticity_true)) / (1.1e-4)
ellipticity_error_upper_mr = ellipticity_requirement_upper_mr - ellipticity_requirement_mr
ellipticity_error_lower_mr = ellipticity_requirement_mr - ellipticity_requirement_lower_mr

most_likely_hr, most_probable_hr, lower_limits_hr, upper_limits_hr, lower_error_hr, upper_error_hr, error_pecision_hr = \
    req.setup_multinest_results(model_mapper=model_mapper, input_model=input_model, sigma_limit=sigma_limit,
                                name='Results/Serial_Tight_hr/')

model_rho_hr = np.array([most_probable_hr[0], most_probable_hr[1]])
model_tau_hr = np.array([most_probable_hr[2], most_probable_hr[3]])
delta_ellipticity_model_hr = convert.convert_to_ellipticity(rho=model_rho_hr, tau=model_tau_hr)

upper_rho_hr = np.array([upper_limits_hr[0], upper_limits_hr[1]])
upper_tau_hr = np.array([upper_limits_hr[2], upper_limits_hr[3]])
delta_ellipticity_upper_hr = convert.convert_to_ellipticity(rho=upper_rho_hr, tau=upper_tau_hr)

lower_rho_hr = np.array([lower_limits_hr[0], lower_limits_hr[1]])
lower_tau_hr = np.array([lower_limits_hr[2], lower_limits_hr[3]])
delta_ellipticity_lower_hr = convert.convert_to_ellipticity(rho=lower_rho_hr, tau=lower_tau_hr)

ellipticity_requirement_hr = (np.sum(delta_ellipticity_model_hr) - np.sum(delta_ellipticity_true)) / (1.1e-4)
ellipticity_requirement_upper_hr = (np.sum(delta_ellipticity_upper_hr) - np.sum(delta_ellipticity_true)) / (1.1e-4)
ellipticity_requirement_lower_hr = (np.sum(delta_ellipticity_lower_hr) - np.sum(delta_ellipticity_true)) / (1.1e-4)
ellipticity_error_upper_hr = ellipticity_requirement_upper_hr - ellipticity_requirement_hr
ellipticity_error_lower_hr = ellipticity_requirement_hr - ellipticity_requirement_lower_hr

print('Induced Delta Elliptcitiy / Requirement')
print((np.sum(delta_ellipticity_model_hr) - np.sum(delta_ellipticity_true)) / (1.1e-4))
print((np.sum(delta_ellipticity_upper_hr) - np.sum(delta_ellipticity_true)) / (1.1e-4))
print((np.sum(delta_ellipticity_lower_hr) - np.sum(delta_ellipticity_true)) / (1.1e-4))
print()


most_likely_x2, most_probable_x2, lower_limits_x2, upper_limits_x2, lower_error_x2, upper_error_x2, error_pecision_x2 = \
    req.setup_multinest_results(model_mapper=model_mapper, input_model=input_model, sigma_limit=sigma_limit,
                                name='Results/Serial_Tight_x2/')

model_rho_x2 = np.array([most_probable_x2[0], most_probable_x2[1]])
model_tau_x2 = np.array([most_probable_x2[2], most_probable_x2[3]])
delta_ellipticity_model_x2 = convert.convert_to_ellipticity(rho=model_rho_x2, tau=model_tau_x2)

upper_rho_x2 = np.array([upper_limits_x2[0], upper_limits_x2[1]])
upper_tau_x2 = np.array([upper_limits_x2[2], upper_limits_x2[3]])
delta_ellipticity_upper_x2 = convert.convert_to_ellipticity(rho=upper_rho_x2, tau=upper_tau_x2)

lower_rho_x2 = np.array([lower_limits_x2[0], lower_limits_x2[1]])
lower_tau_x2 = np.array([lower_limits_x2[2], lower_limits_x2[3]])
delta_ellipticity_lower_x2 = convert.convert_to_ellipticity(rho=lower_rho_x2, tau=lower_tau_x2)

ellipticity_requirement_x2 = (np.sum(delta_ellipticity_model_x2) - np.sum(delta_ellipticity_true)) / (1.1e-4)
ellipticity_requirement_upper_x2 = (np.sum(delta_ellipticity_upper_x2) - np.sum(delta_ellipticity_true)) / (1.1e-4)
ellipticity_requirement_lower_x2 = (np.sum(delta_ellipticity_lower_x2) - np.sum(delta_ellipticity_true)) / (1.1e-4)
ellipticity_error_upper_x2 = ellipticity_requirement_upper_x2 - ellipticity_requirement_x2
ellipticity_error_lower_x2 = ellipticity_requirement_x2 - ellipticity_requirement_lower_x2

print('Induced Delta Elliptcitiy / Requirement')
print((np.sum(delta_ellipticity_model_x2) - np.sum(delta_ellipticity_true)) / (1.1e-4))
print((np.sum(delta_ellipticity_upper_x2) - np.sum(delta_ellipticity_true)) / (1.1e-4))
print((np.sum(delta_ellipticity_lower_x2) - np.sum(delta_ellipticity_true)) / (1.1e-4))
print()

plt.plot([496, 993, 1986, 3972], [1.0, 1.0, 1.0, 1.0], linestyle='--', color='r')
plt.plot([496, 993, 1986, 3972], [-1.0, -1.0, -1.0, -1.0], linestyle='--', color='r')
plt.errorbar(x=[496, 993, 1986, 3972], y=[ellipticity_requirement_lr, ellipticity_requirement_mr,
                                   ellipticity_requirement_hr, ellipticity_requirement_x2],
             yerr=[[ellipticity_error_lower_lr, ellipticity_error_lower_mr, 
                    ellipticity_error_lower_hr, ellipticity_error_lower_x2],
                   [ellipticity_error_upper_lr, ellipticity_error_upper_mr,
                    ellipticity_error_upper_hr, ellipticity_error_upper_x2]], color='b', ecolor='b')
plt.xlabel('Number of columns', fontsize=12)
plt.ylabel('Delta Ellipticity / Requirement (1.1e-4)', fontsize=12)
plt.show()


"""
@file python/ELViS/CR.py
@date 07/13/16
@author user
"""

__author__ = 'hudelot'

import sys

import astropy.io.fits as pyfits
import numpy as np
from workspace_jam.scripts.cosmic_rays import fitslib
# import scipy
from scipy.interpolate import InterpolatedUnivariateSpline, interp1d


class CosmicRays():
    """ Handle cosmic rays simulation """

    def __init__(self, log, shape, cr_fluxscaling, cr_length_file, cr_distance_file, seed=-1):
        self.image = None
        self.cr_length_file = cr_length_file
        self.cr_distance_file = cr_distance_file
        self.cr = None
        self.nx = shape[1]
        self.ny = shape[0]
        self.cr_fluxscaling = cr_fluxscaling
        if seed == -1:
            seed = np.random.randint(0, 1e9)  # Use one seed, so all regions have identical column non-uniformity.
        np.random.seed(seed)
        self.log = log

    def set_ifiles(self):

        # Check if files are ASCII or FITS
        if fitslib.isfits(self.cr_length_file):
            try:
                data = pyfits.getdata(self.cr_length_file, extname='DATA')
            except KeyError:
                self.log.error("No DATA extention in FITS file : %s " % self.cr_length_file)
                sys.exit(1)
            crLengths = np.array(data.tolist())
        else:
            crLengths = np.loadtxt(self.cr_length_file)

        if fitslib.isfits(self.cr_distance_file):
            try:
                data = pyfits.getdata(self.cr_distance_file, extname='DATA')
            except KeyError:
                self.log.error("No DATA extention in FITS file : %s " % self.cr_distance_file)
                sys.exit(1)
            crDists = np.array(data.tolist())
        else:
            crDists = np.loadtxt(self.cr_distance_file)

        # Read CR informations (distribution functions)
        self.cr = dict(cr_u=crLengths[:, 0], cr_cdf=crLengths[:, 1], cr_cdfn=np.shape(crLengths)[0], cr_v=crDists[:, 0],
                       cr_cde=crDists[:, 1], cr_cden=np.shape(crDists)[0])

    def cosmicRayIntercepts(self, lum, x0, y0, l, phi, fscale):
        """
        Derive cosmic ray streak intercept points.

        :param lum: luminosities of the cosmic ray tracks
        :param x0: central positions of the cosmic ray tracks in x-direction
        :param y0: central positions of the cosmic ray tracks in y-direction
        :param l: lengths of the cosmic ray tracks
        :param phi: orientation angles of the cosmic ray tracks

        :return: map
        :rtype: nd-array
        """
        # create empty array
        crImage = np.zeros((self.ny, self.nx), dtype=np.float64)

        # x and y shifts
        dx = l * np.cos(phi) / 2.  # beware! 0<phi< pi, dx < 0
        dy = l * np.sin(phi) / 2.
        mskdx = np.abs(dx) < 1e-8
        mskdy = np.abs(dy) < 1e-8
        dx[mskdx] = 0.
        dy[mskdy] = 0.

        # pixels in x-direction
        ilo = np.round(x0.copy() - dx)
        msk = ilo < 0.
        ilo[msk] = 0
        ilo = ilo.astype(np.int)

        ihi = np.round(x0.copy() + dx)
        msk = ihi > self.nx
        ihi[msk] = self.nx
        ihi = ihi.astype(np.int)

        # pixels in y-directions
        jlo = np.round(y0.copy() - dy)
        msk = jlo < 0.
        jlo[msk] = 0
        jlo = jlo.astype(np.int)

        jhi = np.round(y0.copy() + dy)
        msk = jhi > self.ny
        jhi[msk] = self.ny
        jhi = jhi.astype(np.int)

        offending_delta = 1.

        # loop over the individual events
        for i, luminosity in enumerate(lum):
            n = 0  # count the intercepts

            u = []
            x = []
            y = []

            # Compute X intercepts on the pixel grid
            if ilo[i] < ihi[i]:
                for xcoord in range(ilo[i], ihi[i]):
                    ok = (xcoord - x0[i]) / dx[i]
                    if np.abs(ok) <= offending_delta:
                        n += 1
                        u.append(ok)
                        x.append(xcoord)
                        y.append(y0[i] + ok * dy[i])
            else:
                for xcoord in range(ihi[i], ilo[i]):
                    ok = (xcoord - x0[i]) / dx[i]
                    if np.abs(ok) <= offending_delta:
                        n += 1
                        u.append(ok)
                        x.append(xcoord)
                        y.append(y0[i] + ok * dy[i])

            # Compute Y intercepts on the pixel grid
            if jlo[i] < jhi[i]:
                for ycoord in range(jlo[i], jhi[i]):
                    ok = (ycoord - y0[i]) / dy[i]
                    if np.abs(ok) <= offending_delta:
                        n += 1
                        u.append(ok)
                        x.append(x0[i] + ok * dx[i])
                        y.append(ycoord)
            else:
                for ycoord in range(jhi[i], jlo[i]):
                    ok = (ycoord - y0[i]) / dy[i]
                    if np.abs(ok) <= offending_delta:
                        n += 1
                        u.append(ok)
                        x.append(x0[i] + ok * dx[i])
                        y.append(ycoord)

            # check if no intercepts were found
            if n < 1:
                xc = int(np.floor(x0[i]))
                yc = int(np.floor(y0[i]))
                crImage[yc, xc] += luminosity * fscale

            # Find the arguments that sort the intersections along the track
            u = np.asarray(u)
            x = np.asarray(x)
            y = np.asarray(y)

            args = np.argsort(u)

            u = u[args]
            x = x[args]
            y = y[args]

            # Decide which cell each interval traverses, and the path length
            for j in range(1, n - 1):
                w = (u[j + 1] - u[j]) / 2.
                cx = int(1 + np.floor((x[j + 1] + x[j]) / 2.))
                cy = int(1 + np.floor((y[j + 1] + y[j]) / 2.))

                if 0 <= cx < self.nx and 0 <= cy < self.ny:
                    crImage[cy, cx] += (w * luminosity * fscale)

        return crImage

    def _drawSingleEvent(self, limit=1000, cr_n=1):
        """ Generate a bunch of cosmic ray events and return a cosmic ray map
        limit : limiting energy for the cosmic ray event
        cr_n : number of events to include

        TODO : Decide if needed !
        """

        # random variable for each event (for cosmic length)
        luck = np.random.rand(int(np.floor(cr_n)))

        # draw the length of the tracks
        ius = InterpolatedUnivariateSpline(self.cr['cr_cdf'], self.cr['cr_u'])
        self.cr['cr_l'] = ius(luck)

        # set the energy directly to the limit
        self.cr['cr_e'] = np.asarray([limit, ] * cr_n)

        # Choose the properties such as positions and an angle from a random Uniform dist
        cr_x = self.nx * np.random.rand(int(np.floor(cr_n)))
        cr_y = self.ny * np.random.rand(int(np.floor(cr_n)))
        cr_phi = np.pi * np.random.rand(int(np.floor(cr_n)))

        # find the intercepts
        CCD_cr = self.cosmicRayIntercepts(self.cr['cr_e'], cr_x, cr_y, self.cr['cr_l'], cr_phi,
                                          self.cr_fluxscaling)

        # count the covering factor
        area_cr = np.count_nonzero(CCD_cr)
        self.log.info('The cosmic ray covering factor is %i pixels i.e. %.3f per cent'
                      % (area_cr, 100. * area_cr / (self.nx * self.ny)))

        return CCD_cr

    def drawEventsToCoveringFactor(self, coveringFraction=1.4, limit=1000, nx=0, ny=0):
        """ Generate a bunch of cosmic ray events and return a cosmic ray map
            :param limit: limiting energy for the cosmic ray event
            :param coveringFraction : covering fraction of cr over the total number of px (in percent) normalixzed for a 565s
                                exposure time
            :param nx: x size of the map
            :param ny: y size of the map
            :return map, length, energy: map of cosmic rays, list of CR length, list of CR energies
        """

        if nx != 0:
            self.nx = nx
        if ny != 0:
            self.ny = ny

        # Prepare the CR map
        CCD_cr = np.zeros((self.ny, self.nx))

        # how many events to draw at once, too large number leads to exceeding the covering fraction

        cdf = self.cr['cr_cdf']
        ucr = self.cr['cr_u']
        aproxpdf = (cdf[1:] - cdf[0:-1]) / (ucr[1:] - ucr[0:-1])
        averlength = (aproxpdf * ucr[1:]).sum() / aproxpdf.sum()
        cr_tot_guess = coveringFraction / 100. * self.nx * self.ny / averlength

        # allocating for a max. 5% error in cover. fraction, aprox.
        # Notice that the minimum number of events will be one...
        cr_n = max(int(cr_tot_guess * 0.05), 1)

        covering = 0.0
        lengths = []
        energies = []
        cr_tot = 0

        while covering < coveringFraction:

            # pseudo-random numbers taken from a uniform distribution between 0 and 1
            luck = np.random.rand(cr_n)

            # draw the length of the tracks
            ius = interp1d(self.cr['cr_cdf'], self.cr['cr_u'], kind='slinear')
            self.cr['cr_l'] = ius(luck)

            if limit is None:
                ius = interp1d(self.cr['cr_cde'], self.cr['cr_v'], kind='slinear')
                self.cr['cr_e'] = ius(luck)
            else:
                # set the energy directly to the limit
                self.cr['cr_e'] = np.asarray([limit, ])

            lengths += self.cr['cr_l'].tolist()
            energies += self.cr['cr_e'].tolist()

            # Choose the properties such as positions and an angle from a random Uniform dist
            cr_x = self.nx * np.random.rand(int(np.floor(cr_n)))
            cr_y = self.ny * np.random.rand(int(np.floor(cr_n)))
            cr_phi = np.pi * np.random.rand(int(np.floor(cr_n)))

            # find the intercepts
            CCD_cr += self.cosmicRayIntercepts(self.cr['cr_e'], cr_x, cr_y, self.cr['cr_l'], cr_phi,
                                               self.cr_fluxscaling)

            # count the covering factor
            area_cr = np.count_nonzero(CCD_cr)
            covering = 100. * area_cr / (self.nx * self.ny)

            cr_tot += cr_n
            text = 'The cosmic ray covering factor is %i pixels i.e. %.3f per cent (total number of cr : %i)' % (
                area_cr, covering, cr_tot)
            self.log.info(text)

        return CCD_cr



from test.simulation import makers

# Welcome to the PyAutoCTI test suite data maker. Here, we'll make the suite of data that we use to test and profile
# PyAutoCTI. This consists of the following sets of images:

# A uniform charge injectiono line image, generated using a parallel CTI model with 1 trap species.

# Each image is generated at 4 resolutions, 36x36 (Patch), 120x120 (Low_Res), 300x300 (Mid_Res), 600x600 (High_Res).

data_resolutions = ['patch']

# To simulate each lens, we pass it a name and call its maker. In the makers.py file, you'll see these functions.
makers.make_ci_uniform_parallel_x1_species(data_resolutions=data_resolutions, normalizations=[1000.0, 84700.0])
makers.make_ci_uniform_serial_x1_species(data_resolutions=data_resolutions, normalizations=[1000.0, 84700.0])
makers.make_ci_uniform_parallel_and_serial_x1_species(data_resolutions=data_resolutions, normalizations=[1000.0, 84700.0])
makers.make_ci_uniform_parallel_x3_species(data_resolutions=data_resolutions, normalizations=[1000.0, 84700.0])
makers.make_ci_uniform_serial_x3_species(data_resolutions=data_resolutions, normalizations=[1000.0, 84700.0])
makers.make_ci_uniform_parallel_and_serial_x3_species(data_resolutions=data_resolutions, normalizations=[1000.0, 84700.0])
makers.make_ci_uniform_cosmic_rays_parallel_x1_species(data_resolutions=data_resolutions, normalizations=[1000.0, 84700.0])
makers.make_ci_uniform_cosmic_rays_serial_x1_species(data_resolutions=data_resolutions, normalizations=[1000.0, 84700.0])
makers.make_ci_uniform_cosmic_rays_parallel_and_serial_x1_species(data_resolutions=data_resolutions,
                                                                  normalizations=[1000.0, 84700.0])


import numpy as np

from autocti.charge_injection import ci_hyper

class TestScaledNoise:

    def test__noise_scaling_map_all_0s__scaled_noise_is_0s(self):
        noise_scaling_map = np.array([[0.0, 0.0],
                                      [0.0, 0.0]])

        hyper = ci_hyper.CIHyperNoiseScalar(scale_factor=10.0)
        scaled_noise_map = hyper.scaled_noise_map_from_noise_scaling(noise_scaling_map)

        assert (scaled_noise_map == np.array([[0.0, 0.0],
                                              [0.0, 0.0]])).all()

    def test__noise_scaling_map_has_values__scaled_noise_is_baseline_noise_plus_values_times_noise_factor(self):
        noise_scaling_map = np.array([[1.0, 2.0],
                                      [5.0, 3.0]])

        hyper = ci_hyper.CIHyperNoiseScalar(scale_factor=10.0)
        scaled_noise_map = hyper.scaled_noise_map_from_noise_scaling(noise_scaling_map)

        assert (scaled_noise_map == np.array([[10.0, 20.0],
                                              [50.0, 30.0]])).all()



def phase_tag_from_phase_settings(columns, rows,
                                  parallel_front_edge_mask_rows, parallel_trails_mask_rows,
                                  serial_front_edge_mask_columns, serial_trails_mask_columns,
                                  parallel_total_density_range, serial_total_density_range,
                                  cosmic_ray_parallel_buffer,  cosmic_ray_serial_buffer, cosmic_ray_diagonal_buffer):

    columns_tag = columns_tag_from_columns(columns=columns)
    rows_tag = rows_tag_from_rows(rows=rows)

    parallel_front_edge_mask_rows_tag = parallel_front_edge_mask_rows_tag_from_parallel_front_edge_mask_rows(
        parallel_front_edge_mask_rows=parallel_front_edge_mask_rows)
    parallel_trails_mask_rows_tag = parallel_trails_mask_rows_tag_from_parallel_trails_mask_rows(
        parallel_trails_mask_rows=parallel_trails_mask_rows)
    serial_front_edge_mask_columns_tag = serial_front_edge_mask_columns_tag_from_serial_front_edge_mask_columns(
        serial_front_edge_mask_columns=serial_front_edge_mask_columns)
    serial_trails_mask_columns_tag = serial_trails_mask_columns_tag_from_serial_trails_mask_columns(
        serial_trails_mask_columns=serial_trails_mask_columns)

    parallel_total_density_range_tag = parallel_total_density_range_tag_from_parallel_total_density_range(
        parallel_total_density_range=parallel_total_density_range)
    serial_total_density_range_tag = serial_total_density_range_tag_from_serial_total_density_range(
        serial_total_density_range=serial_total_density_range)

    cosmic_ray_buffer_tag = cosmic_ray_buffer_tag_from_cosmic_ray_buffers(
        cosmic_ray_parallel_buffer=cosmic_ray_parallel_buffer, cosmic_ray_serial_buffer=cosmic_ray_serial_buffer,
        cosmic_ray_diagonal_buffer=cosmic_ray_diagonal_buffer)

    return columns_tag + \
           rows_tag + \
           parallel_front_edge_mask_rows_tag + \
           parallel_trails_mask_rows_tag + \
           serial_front_edge_mask_columns_tag + \
           serial_trails_mask_columns_tag + \
           parallel_total_density_range_tag + \
           serial_total_density_range_tag + \
           cosmic_ray_buffer_tag

def columns_tag_from_columns(columns):
    """Generate a columns tag, to customize phase names based on the number of columns of data extracted in the fit,
    which is used to speed up parallel CTI fits.

    This changes the phase name 'phase_name' as follows:

    columns = None -> phase_name
    columns = 10 -> phase_name_col_10
    columns = 60 -> phase_name_col_60
    """
    if columns == None:
        return ''
    else:
        return '_col_' + str(int(columns))
    
def rows_tag_from_rows(rows):
    """Generate a rows tag, to customize phase names based on the number of rows of data extracted in the fit,
    which is used to speed up serial CTI fits.

    This changes the phase name 'phase_name' as follows:

    rows = None -> phase_name
    rows = (0, 10) -> phase_name_rows_(0,10)
    rows = (20, 60) -> phase_name_rows_(20,60)
    """
    if rows == None:
        return ''
    else:
        x0 = str(rows[0])
        x1 = str(rows[1])
        return ('_rows_(' + x0 + ',' + x1 + ')')

def parallel_front_edge_mask_rows_tag_from_parallel_front_edge_mask_rows(parallel_front_edge_mask_rows):
    """Generate a parallel_front_edge_mask_rows tag, to customize phase names based on the number of rows in the charge
    injection region at the front edge of the parallel clocking direction are masked during the fit,

    This changes the phase name 'phase_name' as follows:

    parallel_front_edge_mask_rows = None -> phase_name
    parallel_front_edge_mask_rows = (0, 10) -> phase_name_parallel_front_edge_mask_rows_(0,10)
    parallel_front_edge_mask_rows = (20, 60) -> phase_name_parallel_front_edge_mask_rows_(20,60)
    """
    if parallel_front_edge_mask_rows == None:
        return ''
    else:
        x0 = str(parallel_front_edge_mask_rows[0])
        x1 = str(parallel_front_edge_mask_rows[1])
        return ('_par_front_mask_rows_(' + x0 + ',' + x1 + ')')


def parallel_trails_mask_rows_tag_from_parallel_trails_mask_rows(parallel_trails_mask_rows):
    """Generate a parallel_trails_mask_rows tag, to customize phase names based on the number of rows in the charge
    injection region in the trails of the parallel clocking direction are masked during the fit,

    This changes the phase name 'phase_name' as follows:

    parallel_trails_mask_rows = None -> phase_name
    parallel_trails_mask_rows = (0, 10) -> phase_name_parallel_trails_mask_rows_(0,10)
    parallel_trails_mask_rows = (20, 60) -> phase_name_parallel_trails_mask_rows_(20,60)
    """
    if parallel_trails_mask_rows == None:
        return ''
    else:
        x0 = str(parallel_trails_mask_rows[0])
        x1 = str(parallel_trails_mask_rows[1])
        return ('_par_trails_mask_rows_(' + x0 + ',' + x1 + ')')


def serial_front_edge_mask_columns_tag_from_serial_front_edge_mask_columns(serial_front_edge_mask_columns):
    """Generate a serial_front_edge_mask_columns tag, to customize phase names based on the number of columns in the
    charge  injection region at the front edge of the serial clocking direction are masked during the fit,

    This changes the phase name 'phase_name' as follows:

    serial_front_edge_mask_columns = None -> phase_name
    serial_front_edge_mask_columns = (0, 10) -> phase_name_serial_front_edge_mask_columns_(0,10)
    serial_front_edge_mask_columns = (20, 60) -> phase_name_serial_front_edge_mask_columns_(20,60)
    """
    if serial_front_edge_mask_columns == None:
        return ''
    else:
        x0 = str(serial_front_edge_mask_columns[0])
        x1 = str(serial_front_edge_mask_columns[1])
        return ('_ser_front_mask_col_(' + x0 + ',' + x1 + ')')


def serial_trails_mask_columns_tag_from_serial_trails_mask_columns(serial_trails_mask_columns):
    """Generate a serial_trails_mask_columns tag, to customize phase names based on the number of columns in the charge
    injection region in the trails of the serial clocking direction are masked during the fit,

    This changes the phase name 'phase_name' as follows:

    serial_trails_mask_columns = None -> phase_name
    serial_trails_mask_columns = (0, 10) -> phase_name_serial_trails_mask_columns_(0,10)
    serial_trails_mask_columns = (20, 60) -> phase_name_serial_trails_mask_columns_(20,60)
    """
    if serial_trails_mask_columns == None:
        return ''
    else:
        x0 = str(serial_trails_mask_columns[0])
        x1 = str(serial_trails_mask_columns[1])
        return ('_ser_trails_mask_col_(' + x0 + ',' + x1 + ')')

def parallel_total_density_range_tag_from_parallel_total_density_range(parallel_total_density_range):
    """Generate a parallel_total_density_range tag, to customize phase names based on the range of values in total \
    density that are allowed for the non-linear search in the parallel direction.

    This changes the phase name 'phase_name' as follows:

    parallel_total_density_range = None -> phase_name
    parallel_total_density_range = (0, 10) -> phase_name_parallel_total_density_range_(0,10)
    parallel_total_density_range = (20, 60) -> phase_name_parallel_total_density_range_(20,60)
    """
    if parallel_total_density_range == None:
        return ''
    else:
        x0 = str(parallel_total_density_range[0])
        x1 = str(parallel_total_density_range[1])
        return ('_par_range_(' + x0 + ',' + x1 + ')')

def serial_total_density_range_tag_from_serial_total_density_range(serial_total_density_range):
    """Generate a serial_total_density_range tag, to customize phase names based on the range of values in total \
    density that are allowed for the non-linear search in the serial direction.

    This changes the phase name 'phase_name' as follows:

    serial_total_density_range = None -> phase_name
    serial_total_density_range = (0, 10) -> phase_name_serial_total_density_range_(0,10)
    serial_total_density_range = (20, 60) -> phase_name_serial_total_density_range_(20,60)
    """
    if serial_total_density_range == None:
        return ''
    else:
        x0 = str(serial_total_density_range[0])
        x1 = str(serial_total_density_range[1])
        return ('_ser_range_(' + x0 + ',' + x1 + ')')

def cosmic_ray_buffer_tag_from_cosmic_ray_buffers(cosmic_ray_parallel_buffer, cosmic_ray_serial_buffer,
                                                  cosmic_ray_diagonal_buffer):
    """Generate a cosmic ray buffer tag, to customize phase names based on the size of the cosmic ray masks in the \
    parallel, serial and diagonal directions

    This changes the phase name 'phase_name' as follows:

    cosmic_ray_parallel_buffer = 1, cosmic_ray_serial_buffer=2, cosmic_ray_diagonal_buffer=3 = -> phase_name_cr_p1s2d3
    cosmic_ray_parallel_buffer = 10, cosmic_ray_serial_buffer=5, cosmic_ray_diagonal_buffer=1 = -> phase_name_cr_p10s5d1
    """

    if cosmic_ray_diagonal_buffer is None and cosmic_ray_serial_buffer is None and cosmic_ray_diagonal_buffer is None:
        return ''
    
    if cosmic_ray_parallel_buffer is None:
        cosmic_ray_parallel_buffer_tag = ''
    else:
        cosmic_ray_parallel_buffer_tag = 'p' + str(cosmic_ray_parallel_buffer)
    
    if cosmic_ray_serial_buffer is None:
        cosmic_ray_serial_buffer_tag = ''
    else:
        cosmic_ray_serial_buffer_tag = 's' + str(cosmic_ray_serial_buffer)

    if cosmic_ray_diagonal_buffer is None:
        cosmic_ray_diagonal_buffer_tag = ''
    else:
        cosmic_ray_diagonal_buffer_tag = 'd' + str(cosmic_ray_diagonal_buffer)

    return '_cr_' + cosmic_ray_parallel_buffer_tag + cosmic_ray_serial_buffer_tag + cosmic_ray_diagonal_buffer_tag


import numpy as np

from autocti.charge_injection.plotters import data_plotters
from autocti.data import mask as msk

from test.fixtures import make_plot_patch
import os
import pytest

@pytest.fixture(name='data_plotter_path')
def make_data_plotter_setup():
    return "{}/../../test_files/plotting/data/".format(os.path.dirname(os.path.realpath(__file__)))


@pytest.fixture(name='mask')
def make_mask():
    return msk.Mask.empty_for_shape(shape=(6, 6))


@pytest.fixture(name='image')
def make_image():
    return np.ones((6, 6))


@pytest.fixture(name='noise_map')
def make_noise_map():
    return 2.0 * np.ones((6, 6))


@pytest.fixture(name='ci_pre_cti')
def make_ci_pre_cti():
    return 3.0 * np.ones((6, 6))


def test__image_is_output(image, mask, data_plotter_path, plot_patch):
    data_plotters.plot_image(image=image, mask=mask, extract_array_from_mask=True,
                             cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                             output_path=data_plotter_path,
                             output_format='png')
    assert data_plotter_path + 'image.png' in plot_patch.paths


def test__noise_map_is_output(noise_map, mask, data_plotter_path, plot_patch):
    data_plotters.plot_noise_map(noise_map=noise_map, mask=mask, extract_array_from_mask=True,
                                 cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                 output_path=data_plotter_path, output_format='png')
    assert data_plotter_path + 'noise_map.png' in plot_patch.paths


def test__ci_pre_cti_is_output(ci_pre_cti, mask, data_plotter_path, plot_patch):
    data_plotters.plot_ci_pre_cti(ci_pre_cti=ci_pre_cti, mask=mask, extract_array_from_mask=True,
                                  cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                  output_path=data_plotter_path, output_format='png')
    assert data_plotter_path + 'ci_pre_cti.png' in plot_patch.paths


def test__signal_to_noise_map_is_output(image, noise_map, mask, data_plotter_path, plot_patch):
    data_plotters.plot_signal_to_noise_map(signal_to_noise_map=image / noise_map, mask=mask,
                                           extract_array_from_mask=True, cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                           output_path=data_plotter_path,
                                           output_format='png')
    assert data_plotter_path + 'signal_to_noise_map.png' in plot_patch.paths



#
# Copyright (C) 2012-2020 Euclid Science Ground Segment
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3.0 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
File: python/SHE_ArCTIC/ArcticParams.pyeated on: 02/13/18
Author: James Nightingale
"""

import numpy as np


class ArcticParams(object):

    def __init__(self, parallel_ccd=None, serial_ccd=None, parallel_species=None, serial_species=None):
        """Sets up the arctic CTI model using parallel and serial parameters specified using a child of the
        ArcticParams.ParallelParams and ArcticParams.SerialParams abstract base classes.

        Parameters
        ----------
        parallel_ccd: CCD
            Class describing the state of the CCD in the parallel direction
        serial_ccd: CCD
            Class describing the state of the CCD in the serial direction
        parallel_species : [ArcticParams.ParallelParams]
           The parallel parameters for the arctic CTI model
        serial_species : [ArcticParams.SerialParams]
           The serial parameters for the arctic CTI model
        """
        self.parallel_ccd = parallel_ccd
        self.serial_ccd = serial_ccd
        self.parallel_species = parallel_species or []
        self.serial_species = serial_species or []

    def update_fits_header_info(self, ext_header):
        """Output the CTI model parameters into the fits header of a fits image.

        Parameters
        -----------
        ext_header : astropy.io.hdulist
            The opened header of the astropy fits header.
        """

        def add_species(name, species_list):
            for i, species in species_list:
                ext_header.set('cte_pt{}d'.format(i), species.trap_density,
                               'Trap species {} density ({})'.format(i, name))
                ext_header.set('cte_pt{}t'.format(i), species.trap_lifetime,
                               'Trap species {} lifetime ({})'.format(i, name))

        add_species("Parallel", self.parallel_species)
        add_species("Serial", self.serial_species)

        if self.serial_ccd is not None:
            ext_header.set('cte_swln', self.serial_ccd.well_notch_depth, 'CCD Well notch depth (Serial)')
            ext_header.set('cte_swlp', self.serial_ccd.well_fill_beta, 'CCD Well filling power (Serial)')

        if self.parallel_ccd is not None:
            ext_header.set('cte_pwln', self.parallel_ccd.well_notch_depth, 'CCD Well notch depth (Parallel)')
            ext_header.set('cte_pwlp', self.parallel_ccd.well_fill_beta, 'CCD Well filling power (Parallel)')

        return ext_header


class CCD(object):

    def __init__(self, well_notch_depth=1e-9, well_fill_alpha=1.0, well_fill_beta=0.58, well_fill_gamma=0.0):
        """Abstract base class of the cti model parameters. Parameters associated with the traps are set via a child \
        class.

        Parameters
        ----------
        well_notch_depth : float
            The CCD notch depth
        well_fill_alpha : float
            The volume-filling coefficient (alpha) of how an electron cloud fills the volume of a pixel.
        well_fill_beta : float
            The volume-filling power (beta) of how an electron cloud fills the volume of a pixel.
        well_fill_gamma : float
            The volume-filling constant (gamma) of how an electron cloud fills the volume of a pixel.
        """
        self.well_notch_depth = well_notch_depth
        self.well_fill_alpha = well_fill_alpha
        self.well_fill_beta = well_fill_beta
        self.well_fill_gamma = well_fill_gamma

    def __repr__(self):
        return '\n'.join(('Well Notch Depth: {}'.format(self.well_notch_depth),
                          'Well Fill Alpha: {}'.format(self.well_fill_alpha),
                          'Well Fill Beta: {}'.format(self.well_fill_beta),
                          'Well Fill Gamma: {}'.format(self.well_fill_gamma)))


class Species(object):

    def __init__(self, trap_density=0.13, trap_lifetime=0.25):
        """The CTI model parameters used for parallel clocking, using one species of trap.

        Parameters
        ----------
        trap_density : float
            The trap density of the species.
        trap_lifetime : float
            The trap lifetimes of the species.
        """
        self.trap_density = trap_density
        self.trap_lifetime = trap_lifetime

    def __repr__(self):
        return "\n".join(('Trap Density: {}'.format(self.trap_density),
                          'Trap Lifetime: {}'.format(self.trap_lifetime)))

    @classmethod
    def poisson_species(cls, species, shape, seed=0):
        """For a set of traps with a given set of densities (which are in traps per pixel), compute a new set of \
        trap densities by drawing new values for from a Poisson distribution.

        This requires us to first convert each trap density to the total number of traps in the column.

        This is used to model the random distribution of traps on a CCD, which changes the number of traps in each \
        column.

        Parameters
        -----------
        species
        shape : (int, int)
            The shape of the image, so that the correct number of trap densities are computed.
        seed : int
            The seed of the Poisson random number generator.
        """
        np.random.seed(seed)
        total_traps = tuple(map(lambda sp: sp.trap_density * shape[0], species))
        poisson_densities = [np.random.poisson(total_traps) / shape[0] for _ in range(shape[1])]
        poisson_species = []
        for densities in poisson_densities:
            for i, s in enumerate(species):
                poisson_species.append(Species(trap_density=densities[i], trap_lifetime=s.trap_lifetime))

        return poisson_species



#
# Copyright (C) 2012-2020 Euclid Science Ground Segment
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3.0 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
File: tests/python/CTICIData_test.py

Created on: 02/14/18
Author: user
"""

import numpy as np

from autocti.charge_injection import ci_frame
from autocti.data import mask as msk


class MockPattern(object):

    def __init__(self):
        pass


class TestMaskRemoveRegions:

    def test__remove_one_region(self):

        mask = msk.Mask.from_masked_regions(shape=(3, 3), frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(),
                                            masked_regions=[(0, 3, 2, 3)])

        assert (mask == np.array([[False, False, True],
                                  [False, False, True],
                                  [False, False, True]])).all()

    def test__remove_two_regions(self):

        mask = msk.Mask.from_masked_regions(shape=(3, 3), frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(),
                                            masked_regions=[(0, 3, 2, 3), (0, 2, 0, 2)])

        assert (mask == np.array([[True, True, True],
                                  [True, True, True],
                                  [False, False, True]])).all()


class TestCosmicRayMask:

    def test__cosmic_ray_mask_included_in_total_mask(self):

        cosmic_ray_image = np.array([[False, False, False],
                                [False, True, False],
                                [False, False, False]])

        mask = msk.Mask.from_cosmic_ray_image(shape=(3, 3), frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(),
                                              cosmic_ray_image=cosmic_ray_image)

        assert (mask == np.array([[False, False, False],
                                  [False, True, False],
                                  [False, False, False]])).all()


class TestMaskCosmicsBottomLeftGeometry:

    def test__mask_one_cosmic_ray_with_parallel_mask(self):
        cosmic_ray_image = np.array([[False, False, False],
                                [False, True, False],
                                [False, False, False]])

        mask = msk.Mask.from_cosmic_ray_image(shape=(3, 3), frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(),
                                              cosmic_ray_image=cosmic_ray_image, cosmic_ray_parallel_buffer=1)

        assert (mask == np.array([[False, False, False],
                                  [False, True, False],
                                  [False, True, False]])).all()

    def test__mask_one_cosmic_ray_with_longer_parallel_mask(self):
        cosmic_ray_image = np.array([[False, True, False],
                                [False, False, False],
                                [False, False, False]])

        mask = msk.Mask.from_cosmic_ray_image(shape=(3, 3), frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(),
                                              cosmic_ray_image=cosmic_ray_image, cosmic_ray_parallel_buffer=2)

        assert (mask == np.array([[False, True, False],
                                  [False, True, False],
                                  [False, True, False]])).all()

    def test__mask_one_cosmic_ray_with_serial_mask(self):
        cosmic_ray_image = np.array([[False, False, False],
                                [False, True, False],
                                [False, False, False]])

        mask = msk.Mask.from_cosmic_ray_image(shape=(3, 3), frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(),
                                              cosmic_ray_image=cosmic_ray_image, cosmic_ray_serial_buffer=1)

        assert (mask == np.array([[False, False, False],
                                  [False, True, True],
                                  [False, False, False]])).all()

    def test__mask_one_cosmic_ray_with_longer_serial_mask(self):
        cosmic_ray_image = np.array([[False, False, False],
                                [True, False, False],
                                [False, False, False]])

        mask = msk.Mask.from_cosmic_ray_image(shape=(3, 3), frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(),
                                              cosmic_ray_image=cosmic_ray_image, cosmic_ray_serial_buffer=2)

        assert (mask == np.array([[False, False, False],
                                  [True, True, True],
                                  [False, False, False]])).all()

    def test__mask_one_cosmic_ray_with_diagonal_mask(self):
        cosmic_ray_image = np.array([[False, False, False],
                                [False, True, False],
                                [False, False, False]])

        mask = msk.Mask.from_cosmic_ray_image(shape=(3, 3), frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(),
                                              cosmic_ray_image=cosmic_ray_image, cosmic_ray_diagonal_buffer=1)

        assert (mask == np.array([[False, False, False],
                                  [False, True, True],
                                  [False, True, True]])).all()

    def test__mask_one_cosmic_ray_with_bigger_diagonal_mask(self):
        cosmic_ray_image = np.array([[False, False, False, False],
                                [False, True, False, False],
                                [False, False, False, False],
                                [False, False, False, False]])

        mask = msk.Mask.from_cosmic_ray_image(shape=(4, 4), frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(),
                                              cosmic_ray_image=cosmic_ray_image, cosmic_ray_diagonal_buffer=2)

        assert (mask == np.array([[False, False, False, False],
                                  [False, True, True, True],
                                  [False, True, True, True],
                                  [False, True, True, True]])).all()


class TestMaskCosmicsBottomRightGeometry:

    def test__mask_one_cosmic_ray_with_parallel_mask(self):
        cosmic_ray_image = np.array([[False, False, False],
                                [False, True, False],
                                [False, False, False]])

        mask = msk.Mask.from_cosmic_ray_image(shape=(3, 3), frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(),
                                              cosmic_ray_image=cosmic_ray_image, cosmic_ray_parallel_buffer=1)

        assert (mask == np.array([[False, False, False],
                                  [False, True, False],
                                  [False, True, False]])).all()

    def test__mask_one_cosmic_ray_with_longer_parallel_mask(self):
        cosmic_ray_image = np.array([[False, True, False],
                                [False, False, False],
                                [False, False, False]])

        mask = msk.Mask.from_cosmic_ray_image(shape=(3, 3), frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(),
                                              cosmic_ray_image=cosmic_ray_image, cosmic_ray_parallel_buffer=2)

        assert (mask == np.array([[False, True, False],
                                  [False, True, False],
                                  [False, True, False]])).all()

    def test__mask_one_cosmic_ray_with_serial_mask(self):
        cosmic_ray_image = np.array([[False, False, False],
                                [False, True, False],
                                [False, False, False]])

        mask = msk.Mask.from_cosmic_ray_image(shape=(3, 3), frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(),
                                              cosmic_ray_image=cosmic_ray_image, cosmic_ray_serial_buffer=1)

        assert (mask == np.array([[False, False, False],
                                  [True, True, False],
                                  [False, False, False]])).all()

    def test__mask_one_cosmic_ray_with_longer_serial_mask(self):
        cosmic_ray_image = np.array([[False, False, False],
                                [False, False, True],
                                [False, False, False]])

        mask = msk.Mask.from_cosmic_ray_image(shape=(3, 3), frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(),
                                              cosmic_ray_image=cosmic_ray_image, cosmic_ray_serial_buffer=2)

        assert (mask == np.array([[False, False, False],
                                  [True, True, True],
                                  [False, False, False]])).all()

    def test__mask_one_cosmic_ray_with_diagonal_mask(self):
        cosmic_ray_image = np.array([[False, False, False],
                                [False, True, False],
                                [False, False, False]])

        mask = msk.Mask.from_cosmic_ray_image(shape=(3, 3), frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(),
                                              cosmic_ray_image=cosmic_ray_image, cosmic_ray_diagonal_buffer=1)

        assert (mask == np.array([[False, False, False],
                                  [True, True, False],
                                  [True, True, False]])).all()

    def test__mask_one_cosmic_ray_with_bigger_diagonal_mask(self):
        cosmic_ray_image = np.array([[False, False, False, False],
                                [False, False, False, True],
                                [False, False, False, False],
                                [False, False, False, False]])

        mask = msk.Mask.from_cosmic_ray_image(shape=(4, 4), frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(),
                                              cosmic_ray_image=cosmic_ray_image, cosmic_ray_diagonal_buffer=2)

        assert (mask == np.array([[False, False, False, False],
                                  [False, True, True, True],
                                  [False, True, True, True],
                                  [False, True, True, True]])).all()


class TestMaskCosmicsTopLeftGeometry:

    def test__mask_one_cosmic_ray_with_parallel_mask(self):
        cosmic_ray_image = np.array([[False, False, False],
                                [False, True, False],
                                [False, False, False]])

        mask = msk.Mask.from_cosmic_ray_image(shape=(3, 3), frame_geometry=ci_frame.QuadGeometryEuclid.top_left(),
                                              cosmic_ray_image=cosmic_ray_image, cosmic_ray_parallel_buffer=1)

        assert (mask == np.array([[False, True, False],
                                  [False, True, False],
                                  [False, False, False]])).all()

    def test__mask_one_cosmic_ray_with_longer_parallel_mask(self):
        cosmic_ray_image = np.array([[False, False, False],
                                [False, False, False],
                                [False, True, False]])

        mask = msk.Mask.from_cosmic_ray_image(shape=(3, 3), frame_geometry=ci_frame.QuadGeometryEuclid.top_left(),
                                              cosmic_ray_image=cosmic_ray_image, cosmic_ray_parallel_buffer=2)

        assert (mask == np.array([[False, True, False],
                                  [False, True, False],
                                  [False, True, False]])).all()

    def test__mask_one_cosmic_ray_with_serial_mask(self):
        cosmic_ray_image = np.array([[False, False, False],
                                [False, True, False],
                                [False, False, False]])

        mask = msk.Mask.from_cosmic_ray_image(shape=(3, 3), frame_geometry=ci_frame.QuadGeometryEuclid.top_left(),
                                              cosmic_ray_image=cosmic_ray_image, cosmic_ray_serial_buffer=1)

        assert (mask == np.array([[False, False, False],
                                  [False, True, True],
                                  [False, False, False]])).all()

    def test__mask_one_cosmic_ray_with_longer_serial_mask(self):
        cosmic_ray_image = np.array([[False, False, False],
                                [True, False, False],
                                [False, False, False]])

        mask = msk.Mask.from_cosmic_ray_image(shape=(3, 3), frame_geometry=ci_frame.QuadGeometryEuclid.top_left(),
                                              cosmic_ray_image=cosmic_ray_image, cosmic_ray_serial_buffer=2)

        assert (mask == np.array([[False, False, False],
                                  [True, True, True],
                                  [False, False, False]])).all()

    def test__mask_one_cosmic_ray_with_diagonal_mask(self):
        cosmic_ray_image = np.array([[False, False, False],
                                [False, True, False],
                                [False, False, False]])

        mask = msk.Mask.from_cosmic_ray_image(shape=(3, 3), frame_geometry=ci_frame.QuadGeometryEuclid.top_left(),
                                              cosmic_ray_image=cosmic_ray_image, cosmic_ray_diagonal_buffer=1)

        assert (mask == np.array([[False, True, True],
                                  [False, True, True],
                                  [False, False, False]])).all()

    def test__mask_one_cosmic_ray_with_bigger_diagonal_mask(self):
        cosmic_ray_image = np.array([[False, False, False, False],
                                [False, False, False, False],
                                [False, False, False, False],
                                [False, True, False, False]])

        mask = msk.Mask.from_cosmic_ray_image(shape=(4, 4), frame_geometry=ci_frame.QuadGeometryEuclid.top_left(),
                                              cosmic_ray_image=cosmic_ray_image, cosmic_ray_diagonal_buffer=2)

        assert (mask == np.array([[False, False, False, False],
                                  [False, True, True, True],
                                  [False, True, True, True],
                                  [False, True, True, True]])).all()


class TestMaskCosmicsTopRightGeometry:

    def test__mask_one_cosmic_ray_with_parallel_mask(self):
        cosmic_ray_image = np.array([[False, False, False],
                                [False, True, False],
                                [False, False, False]])

        mask = msk.Mask.from_cosmic_ray_image(shape=(3, 3), frame_geometry=ci_frame.QuadGeometryEuclid.top_right(),
                                              cosmic_ray_image=cosmic_ray_image, cosmic_ray_parallel_buffer=1)

        assert (mask == np.array([[False, True, False],
                                  [False, True, False],
                                  [False, False, False]])).all()

    def test__mask_one_cosmic_ray_with_longer_parallel_mask(self):
        cosmic_ray_image = np.array([[False, False, False],
                                [False, False, False],
                                [False, True, False]])

        mask = msk.Mask.from_cosmic_ray_image(shape=(3, 3), frame_geometry=ci_frame.QuadGeometryEuclid.top_right(),
                                              cosmic_ray_image=cosmic_ray_image, cosmic_ray_parallel_buffer=2)
        assert (mask == np.array([[False, True, False],
                                  [False, True, False],
                                  [False, True, False]])).all()

    def test__mask_one_cosmic_ray_with_serial_mask(self):
        cosmic_ray_image = np.array([[False, False, False],
                                [False, True, False],
                                [False, False, False]])

        mask = msk.Mask.from_cosmic_ray_image(shape=(3, 3), frame_geometry=ci_frame.QuadGeometryEuclid.top_right(),
                                              cosmic_ray_image=cosmic_ray_image, cosmic_ray_serial_buffer=1)

        assert (mask == np.array([[False, False, False],
                                  [True, True, False],
                                  [False, False, False]])).all()

    def test__mask_one_cosmic_ray_with_longer_serial_mask(self):
        cosmic_ray_image = np.array([[False, False, False],
                                [False, False, True],
                                [False, False, False]])

        mask = msk.Mask.from_cosmic_ray_image(shape=(3, 3), frame_geometry=ci_frame.QuadGeometryEuclid.top_right(),
                                              cosmic_ray_image=cosmic_ray_image, cosmic_ray_serial_buffer=2)

        assert (mask == np.array([[False, False, False],
                                  [True, True, True],
                                  [False, False, False]])).all()

    def test__mask_one_cosmic_ray_with_diagonal_mask(self):
        cosmic_ray_image = np.array([[False, False, False],
                                [False, True, False],
                                [False, False, False]])

        mask = msk.Mask.from_cosmic_ray_image(shape=(3, 3), frame_geometry=ci_frame.QuadGeometryEuclid.top_right(),
                                              cosmic_ray_image=cosmic_ray_image, cosmic_ray_diagonal_buffer=1)

        assert (mask == np.array([[True, True, False],
                                  [True, True, False],
                                  [False, False, False]])).all()

    def test__mask_one_cosmic_ray_with_bigger_diagonal_mask(self):
        cosmic_ray_image = np.array([[False, False, False, False],
                                [False, False, False, False],
                                [False, False, False, False],
                                [False, False, False, True]])

        mask = msk.Mask.from_cosmic_ray_image(shape=(4, 4), frame_geometry=ci_frame.QuadGeometryEuclid.top_right(),
                                              cosmic_ray_image=cosmic_ray_image, cosmic_ray_diagonal_buffer=2)

        assert (mask == np.array([[False, False, False, False],
                                  [False, True, True, True],
                                  [False, True, True, True],
                                  [False, True, True, True]])).all()



from autofit.tools import path_util
from autofit.optimize import non_linear as nl
from autofit.mapper import prior
from autofit.mapper import prior_model
from autocti.data import mask as msk
from autocti.pipeline import pipeline as pl
from autocti.pipeline import phase as ph
from autocti.model import arctic_params

# In this pipeline, we'll perform an analysis which fits two serial trap species to a set of charge
# injection imaging data. This will include a hyper-phase which scales the noise in the analysis, to prevent
# over-fitting the highest S/N charge injection images. The pipeline uses four phases:

# Phase 1) Fit a small section (60 columns of every charge injection) using a serial CTI model
#          with 1 trap species and a model for the serial CCD volume filling parameters.

# Phase 2) Fit a small section (again, 60 columns) using a serial CTI model with 2 trap species and a model for the
#          serial CCD volume filling parameters. The priors on trap densities and volume filling parameters are
#          initialized from the results of phase 1.

# Phase 3) Use the best-fit model from phase 2 to scale the noise of each image, to ensure that the higher and
#          lower S/N images are weighted more equally in their contribution to the likelihood.

# Phase 4) Refit the phase 2 model, using priors initialized from the results of phase 2 and the scaled noise-map
#          computed in phase 3.

def make_pipeline(phase_folders=None, tag_phases=True, mask_function=msk.Mask.empty_for_shape,
                  serial_front_edge_mask_columns=None, serial_trails_mask_columns=None,
                  serial_total_density_range=None,
                  cosmic_ray_parallel_buffer=None, cosmic_ray_serial_buffer=None, cosmic_ray_diagonal_buffer=None):

    pipeline_name = 'pipeline_normal__serial_x2'
    
    # This function uses the phase folders and pipeline name to set up the output directory structure,
    # e.g. 'autolens_workspace/output/phase_folder_1/phase_folder_2/pipeline_name/phase_name/'
    phase_folders = path_util.phase_folders_from_phase_folders_and_pipeline_name(phase_folders=phase_folders,
                                                                                pipeline_name=pipeline_name)

    ### PHASE 1 ###

    # In phase 1, we will fit the data with a 3 species serial CTI model and serial CCD filling model. In this
    # phase we will:

    # 1) Use the complete charge injection image, as opposed to extracting a sub-set of columns.
    # 3) Initialize the priors on the serial CTI model from the results of the initialize pipeline's phase 2.

    class SerialPhase(ph.SerialPhase):

        def pass_priors(self, results):

            self.serial_species = results.from_phase('phase_2_serial_x2').variable.serial_species
            self.serial_ccd = results.from_phase('phase_2_serial_x2').variable.serial_ccd
            self.serial_ccd.well_fill_alpha = 1.0
            self.serial_ccd.well_fill_gamma = 0.0

    phase1 = SerialPhase(phase_name='phase_1_serial_x2', phase_folders=phase_folders,
                         tag_phases=tag_phases,
                         serial_species=[prior_model.PriorModel(arctic_params.Species),
                                         prior_model.PriorModel(arctic_params.Species)],
                         serial_ccd=arctic_params.CCD,
                         optimizer_class=nl.MultiNest, mask_function=mask_function,
                         serial_front_edge_mask_columns=serial_front_edge_mask_columns,
                         serial_trails_mask_columns=serial_trails_mask_columns,
                         serial_total_density_range=serial_total_density_range,
                         cosmic_ray_parallel_buffer=cosmic_ray_parallel_buffer,
                         cosmic_ray_serial_buffer=cosmic_ray_serial_buffer,
                         cosmic_ray_diagonal_buffer=cosmic_ray_diagonal_buffer)

    # For the final CTI model, constant efficiency mode has a tendency to sample parameter space too fast and infer an
    # inaccurate model. Thus, we turn it off for phase 2.

    phase1.optimizer.const_efficiency_mode = False
    phase1.optimizer.n_live_points = 50
    phase1.optimizer.sampling_efficiency = 0.3

    return pl.Pipeline(pipeline_name, phase1)


from autofit.tools import path_util
from autocti.charge_injection import ci_data
from autocti.charge_injection import ci_pattern
from autocti.data import util

from autocti.model import arctic_settings
from autocti.model import arctic_params

from workspace_jam.tools.data_makers import tools

import os

# This tool allows one to make simulated charge injection imaging data-sets for calibrating parallel charge transfer
# inefficiency, which can be used to test example pipelines and investigate CTI modeling on data-sets where the
# 'true' answer is known.

# The 'ci_data_type', 'ci_data_model' and 'ci_data_resolution' determine the directory the output data folder, e.g:

# The image will be output as '/workspace/data/ci_data_type/ci_data_model/ci_data_resolution/ci_image.fits'.
# The noise-map will be output as '/workspace/data/ci_data_type/ci_data_model/ci_data_resolution/ci_noise_map.fits'.
# The pre cti ci image will be output as '/workspace/data/ci_data_type/ci_data_model/ci_data_resolution/ci_pre_cti.fits'.

# Lets setup the relative path to the workspace, so we can output the data in the 'data' folder..
workspace_path = '{}/../../../../'.format(os.path.dirname(os.path.realpath(__file__)))

ci_data_type = 'ci_images_non_uniform'
ci_data_model = 'serial_x3'
ci_data_resolution = 'low_resolution'

# Create the path where the data will be output, which in this case is
# '/workspace/data/ci_images_uniform/parallel_x2_species/high_res/'
ci_data_path = path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=['data', ci_data_type, ci_data_model, ci_data_resolution])

# This tools loads the correct shape of the image and frame geometry, for the ci_data_resoultion string.
shape, frame_geometry = \
    tools.serial_shape_and_frame_geometry_from_ci_data_resolution(ci_data_resolution=ci_data_resolution)

# Specify the charge injection regions on the CCD, which in this case is 7 equally spaced rectangular blocks.
ci_regions = [(0, shape[0], 51, shape[1]-20)]

# The normalization of every ci image - this size of this list thus determines how many images are simulated.
normalizations=[100.0, 500.0, 1000.0, 5000.0, 10000.0, 25000.0, 50000.0, 84700.0]

# The CTI settings of arCTIc, which models the CCD read-out including CTI. For serial ci data, we include 'charge
# injection mode' which accounts for the fact that every pixel is transferred over the full CCD.
serial_cti_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=2, n_levels=2000,
                                               charge_injection_mode=False, readout_offset=0)
cti_settings = arctic_settings.ArcticSettings(serial=serial_cti_settings)

# The CTI model parameters of arCTIc, which includes each trap species density / lifetime and the CCD properties for
# serial charge transfer.
serial_species_0 = arctic_params.Species(trap_density=0.01, trap_lifetime=0.8)
serial_species_1 = arctic_params.Species(trap_density=0.03, trap_lifetime=4.0)
serial_species_2 = arctic_params.Species(trap_density=0.9, trap_lifetime=20.0)
serial_ccd = arctic_params.CCD(well_notch_depth=1.0e-4, well_fill_alpha=1.0, well_fill_beta=0.58)
cti_params = arctic_params.ArcticParams(serial_species=[serial_species_0, serial_species_1, serial_species_2],
                                        serial_ccd=serial_ccd)

# This function creates uniform blocks of non-charge injection lines, given an input normalization, regions and
# parameters describing its non-uniform properties.
column_deviations = [100.0]*len(normalizations)
row_slopes = [0.0]*len(normalizations)
ci_patterns = ci_pattern.non_uniform_from_lists(normalizations=normalizations, regions=ci_regions,
                                                row_slopes=row_slopes)

# Use the simulate ci patterns to generate the pre-cti charge injection images.
ci_pre_ctis = list(map(lambda ci_pattern, column_deviation :
                       ci_pattern.simulate_ci_pre_cti(shape=shape, column_deviation=column_deviation,
                                                      maximum_normalization=cti_settings.serial.well_depth),
                       ci_patterns, column_deviations))

# Use every ci pattern to simulate a ci image.
ci_datas = list(map(lambda ci_pre_cti, ci_pattern :
                    ci_data.simulate(ci_pre_cti=ci_pre_cti, frame_geometry=frame_geometry,
                                      ci_pattern=ci_pattern, cti_settings=cti_settings, cti_params=cti_params,
                                      read_noise=4.0),
                    ci_pre_ctis, ci_patterns))

# Now, output every image to the data folder as the filename 'ci_data_normalization.fits'
list(map(lambda ci_data, normalization :
         util.numpy_array_2d_to_fits(array_2d=ci_data.image,
                                     file_path=ci_data_path + 'image_' + str(int(normalization)) + '.fits',
                                     overwrite=True),
         ci_datas, normalizations))

# Output every pre-cti image to the data folder as the filename 'ci_pre_cti_normalization.fits'. This allows the
# calibration pipeline to load these images as the model pre-cti images, which is necessary for non-uniform ci patterns.
list(map(lambda ci_data, normalization :
         util.numpy_array_2d_to_fits(array_2d=ci_data.ci_pre_cti,
                                     file_path=ci_data_path + 'ci_pre_cti_' + str(int(normalization)) + '.fits',
                                     overwrite=True),
         ci_datas, normalizations))


from autofit.tools import path_util
from autocti.charge_injection import ci_data
from autocti.charge_injection import ci_pattern
from autocti.model import arctic_params
from autocti.model import arctic_settings

from test.simulation import simulation_util

from workspace_jam.scripts.cosmic_rays import cosmics

import logging
logger = logging.getLogger(__name__)

import os

workspace_path = '{}/../../workspace/'.format(os.path.dirname(os.path.realpath(__file__)))

def simulate_ci_data_from_ci_normalization_region_and_cti_model(ci_data_type, ci_data_model, ci_data_resolution,
                                                                pattern, cti_params, cti_settings, read_noise=1.0,
                                                                cosmic_ray_image=None):

    shape = simulation_util.shape_from_ci_data_resolution(ci_data_resolution=ci_data_resolution)
    frame_geometry = simulation_util.frame_geometry_from_ci_data_resolution(ci_data_resolution=ci_data_resolution)

    ci_pre_cti = pattern.simulate_ci_pre_cti(shape=shape)

    data = ci_data.simulate(ci_pre_cti=ci_pre_cti, frame_geometry=frame_geometry, ci_pattern=pattern,
                            cti_settings=cti_settings, cti_params=cti_params, read_noise=read_noise,
                            cosmic_ray_image=cosmic_ray_image)

    # Now, lets output this simulated ccd-data to the test/data folder.
    test_path = '{}/../'.format(os.path.dirname(os.path.realpath(__file__)))

    ci_data_path = path_util.make_and_return_path_from_path_and_folder_names(
        path=test_path, folder_names=['data', ci_data_type, ci_data_model, ci_data_resolution])

    normalization = str(int(pattern.normalization))

    ci_data.output_ci_data_to_fits(ci_data=data,
                                   image_path=ci_data_path + 'image_' + normalization + '.fits',
                                   noise_map_path=ci_data_path + 'noise_map_'  + normalization + '.fits',
                                   ci_pre_cti_path=ci_data_path + 'ci_pre_cti_'  + normalization + '.fits',
                                   cosmic_ray_image_path=ci_data_path + 'cosmic_ray_image_' + normalization + '.fits',
                                   overwrite=True)

def make_ci_uniform_parallel_x1_species(data_resolutions, normalizations):

    ci_data_type = 'ci_uniform'
    ci_data_model = 'parallel_x1'

    parallel_species = arctic_params.Species(trap_density=1.0, trap_lifetime=3.0)
    parallel_ccd = arctic_params.CCD(well_notch_depth=0.0, well_fill_alpha=1.0,
                                     well_fill_beta=0.5, well_fill_gamma=0.0)
    cti_params = arctic_params.ArcticParams(parallel_ccd=parallel_ccd, parallel_species=[parallel_species])

    parallel_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=2, n_levels=2000,
                                                 charge_injection_mode=False, readout_offset=0)
    cti_settings = arctic_settings.ArcticSettings(parallel=parallel_settings)

    for data_resolution in data_resolutions:
        for normalization in normalizations:

            ci_regions = simulation_util.ci_regions_from_ci_data_resolution(ci_data_resolution=data_resolution)
            pattern = ci_pattern.CIPatternUniform(normalization=normalization, regions=ci_regions)

            simulate_ci_data_from_ci_normalization_region_and_cti_model(
                ci_data_type=ci_data_type, ci_data_model=ci_data_model, ci_data_resolution=data_resolution,
                pattern=pattern, cti_params=cti_params, cti_settings=cti_settings)
            
def make_ci_uniform_serial_x1_species(data_resolutions, normalizations):

    ci_data_type = 'ci_uniform'
    ci_data_model = 'serial_x1'

    serial_species = arctic_params.Species(trap_density=1.0, trap_lifetime=3.0)
    serial_ccd = arctic_params.CCD(well_notch_depth=0.00, well_fill_alpha=1.0,
                                     well_fill_beta=0.5, well_fill_gamma=0.0)
    cti_params = arctic_params.ArcticParams(serial_ccd=serial_ccd, serial_species=[serial_species])

    serial_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=2, n_levels=2000,
                                                 charge_injection_mode=False, readout_offset=0)
    cti_settings = arctic_settings.ArcticSettings(serial=serial_settings)

    for data_resolution in data_resolutions:
        for normalization in normalizations:

            ci_regions = simulation_util.ci_regions_from_ci_data_resolution(ci_data_resolution=data_resolution)
            pattern = ci_pattern.CIPatternUniform(normalization=normalization, regions=ci_regions)

            simulate_ci_data_from_ci_normalization_region_and_cti_model(
                ci_data_type=ci_data_type, ci_data_model=ci_data_model, ci_data_resolution=data_resolution,
                pattern=pattern, cti_params=cti_params, cti_settings=cti_settings)

def make_ci_uniform_parallel_and_serial_x1_species(data_resolutions, normalizations):

    ci_data_type = 'ci_uniform'
    ci_data_model = 'parallel_and_serial_x1'

    parallel_species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.5)
    parallel_ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0,
                                     well_fill_beta=0.5, well_fill_gamma=0.0)

    serial_species = arctic_params.Species(trap_density=1.0, trap_lifetime=3.0)
    serial_ccd = arctic_params.CCD(well_notch_depth=0.00, well_fill_alpha=1.0,
                                     well_fill_beta=0.5, well_fill_gamma=0.0)

    cti_params = arctic_params.ArcticParams(parallel_ccd=parallel_ccd, parallel_species=[parallel_species],
                                            serial_ccd=serial_ccd, serial_species=[serial_species])

    parallel_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=2, n_levels=2000,
                                                 charge_injection_mode=False, readout_offset=0)

    serial_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=2, n_levels=2000,
                                                 charge_injection_mode=False, readout_offset=0)

    cti_settings = arctic_settings.ArcticSettings(parallel=parallel_settings, serial=serial_settings)

    for data_resolution in data_resolutions:
        for normalization in normalizations:

            ci_regions = simulation_util.ci_regions_from_ci_data_resolution(ci_data_resolution=data_resolution)
            pattern = ci_pattern.CIPatternUniform(normalization=normalization, regions=ci_regions)

            simulate_ci_data_from_ci_normalization_region_and_cti_model(
                ci_data_type=ci_data_type, ci_data_model=ci_data_model, ci_data_resolution=data_resolution,
                pattern=pattern, cti_params=cti_params, cti_settings=cti_settings)

def make_ci_uniform_parallel_x3_species(data_resolutions, normalizations):

    ci_data_type = 'ci_uniform'
    ci_data_model = 'parallel_x3'

    parallel_species_0 = arctic_params.Species(trap_density=0.5, trap_lifetime=2.0)
    parallel_species_1 = arctic_params.Species(trap_density=1.5, trap_lifetime=5.0)
    parallel_species_2 = arctic_params.Species(trap_density=2.5, trap_lifetime=20.0)
    parallel_ccd = arctic_params.CCD(well_notch_depth=0.0, well_fill_alpha=1.0,
                                     well_fill_beta=0.5, well_fill_gamma=0.0)
    cti_params = arctic_params.ArcticParams(parallel_ccd=parallel_ccd,
                                parallel_species=[parallel_species_0, parallel_species_1, parallel_species_2])

    parallel_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=2, n_levels=2000,
                                                 charge_injection_mode=False, readout_offset=0)
    cti_settings = arctic_settings.ArcticSettings(parallel=parallel_settings)

    for data_resolution in data_resolutions:
        for normalization in normalizations:
            ci_regions = simulation_util.ci_regions_from_ci_data_resolution(ci_data_resolution=data_resolution)
            pattern = ci_pattern.CIPatternUniform(normalization=normalization, regions=ci_regions)

            simulate_ci_data_from_ci_normalization_region_and_cti_model(
                ci_data_type=ci_data_type, ci_data_model=ci_data_model, ci_data_resolution=data_resolution,
                pattern=pattern, cti_params=cti_params, cti_settings=cti_settings)

def make_ci_uniform_serial_x3_species(data_resolutions, normalizations):

    ci_data_type = 'ci_uniform'
    ci_data_model = 'serial_x3'

    serial_species_0 = arctic_params.Species(trap_density=0.5, trap_lifetime=2.0)
    serial_species_1 = arctic_params.Species(trap_density=1.5, trap_lifetime=5.0)
    serial_species_2 = arctic_params.Species(trap_density=2.5, trap_lifetime=20.0)
    serial_ccd = arctic_params.CCD(well_notch_depth=0.00, well_fill_alpha=1.0,
                                   well_fill_beta=0.5, well_fill_gamma=0.0)
    cti_params = arctic_params.ArcticParams(serial_ccd=serial_ccd,
                                            serial_species=[serial_species_0, serial_species_1, serial_species_2])

    serial_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=2, n_levels=2000,
                                               charge_injection_mode=False, readout_offset=0)
    cti_settings = arctic_settings.ArcticSettings(serial=serial_settings)

    for data_resolution in data_resolutions:
        for normalization in normalizations:
            ci_regions = simulation_util.ci_regions_from_ci_data_resolution(ci_data_resolution=data_resolution)
            pattern = ci_pattern.CIPatternUniform(normalization=normalization, regions=ci_regions)

            simulate_ci_data_from_ci_normalization_region_and_cti_model(
                ci_data_type=ci_data_type, ci_data_model=ci_data_model, ci_data_resolution=data_resolution,
                pattern=pattern, cti_params=cti_params, cti_settings=cti_settings)

def make_ci_uniform_parallel_and_serial_x3_species(data_resolutions, normalizations):

    ci_data_type = 'ci_uniform'
    ci_data_model = 'parallel_and_serial_x3'

    parallel_species_0 = arctic_params.Species(trap_density=0.5, trap_lifetime=2.0)
    parallel_species_1 = arctic_params.Species(trap_density=1.5, trap_lifetime=5.0)
    parallel_species_2 = arctic_params.Species(trap_density=2.5, trap_lifetime=20.0)
    parallel_ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0,
                                     well_fill_beta=0.5, well_fill_gamma=0.0)

    serial_species_0 = arctic_params.Species(trap_density=0.5, trap_lifetime=2.0)
    serial_species_1 = arctic_params.Species(trap_density=1.5, trap_lifetime=5.0)
    serial_species_2 = arctic_params.Species(trap_density=2.5, trap_lifetime=20.0)
    serial_ccd = arctic_params.CCD(well_notch_depth=0.00, well_fill_alpha=1.0,
                                   well_fill_beta=0.5, well_fill_gamma=0.0)

    cti_params = arctic_params.ArcticParams(parallel_ccd=parallel_ccd,
                                            parallel_species=[parallel_species_0, parallel_species_1,
                                                              parallel_species_2],
                                            serial_ccd=serial_ccd,
                                            serial_species=[serial_species_0, serial_species_1, serial_species_2])

    parallel_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=2, n_levels=2000,
                                                 charge_injection_mode=False, readout_offset=0)

    serial_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=2, n_levels=2000,
                                               charge_injection_mode=False, readout_offset=0)

    cti_settings = arctic_settings.ArcticSettings(parallel=parallel_settings, serial=serial_settings)

    for data_resolution in data_resolutions:
        for normalization in normalizations:
            ci_regions = simulation_util.ci_regions_from_ci_data_resolution(ci_data_resolution=data_resolution)
            pattern = ci_pattern.CIPatternUniform(normalization=normalization, regions=ci_regions)

            simulate_ci_data_from_ci_normalization_region_and_cti_model(
                ci_data_type=ci_data_type, ci_data_model=ci_data_model, ci_data_resolution=data_resolution,
                pattern=pattern, cti_params=cti_params, cti_settings=cti_settings)

def make_ci_uniform_cosmic_rays_parallel_x1_species(data_resolutions, normalizations):

    ci_data_type = 'ci_uniform_cosmic_rays'
    ci_data_model = 'parallel_x1'

    parallel_species = arctic_params.Species(trap_density=1.0, trap_lifetime=3.0)
    parallel_ccd = arctic_params.CCD(well_notch_depth=0.0, well_fill_alpha=1.0,
                                     well_fill_beta=0.5, well_fill_gamma=0.0)
    cti_params = arctic_params.ArcticParams(parallel_ccd=parallel_ccd, parallel_species=[parallel_species])

    parallel_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=2, n_levels=2000,
                                                 charge_injection_mode=False, readout_offset=0)
    cti_settings = arctic_settings.ArcticSettings(parallel=parallel_settings)

    for data_resolution in data_resolutions:

        shape = simulation_util.shape_from_ci_data_resolution(ci_data_resolution=data_resolution)

        for normalization in normalizations:

            ci_regions = simulation_util.ci_regions_from_ci_data_resolution(ci_data_resolution=data_resolution)
            pattern = ci_pattern.CIPatternUniform(normalization=normalization, regions=ci_regions)

            cosmic_ray_image = cosmic_ray_image_from_shape_and_well_depth(shape=shape,
                                                                          well_depth=cti_settings.parallel.well_depth)

            simulate_ci_data_from_ci_normalization_region_and_cti_model(
                ci_data_type=ci_data_type, ci_data_model=ci_data_model, ci_data_resolution=data_resolution,
                pattern=pattern, cti_params=cti_params, cti_settings=cti_settings, cosmic_ray_image=cosmic_ray_image)

def make_ci_uniform_cosmic_rays_serial_x1_species(data_resolutions, normalizations):

    ci_data_type = 'ci_uniform_cosmic_rays'
    ci_data_model = 'serial_x1'

    serial_species = arctic_params.Species(trap_density=1.0, trap_lifetime=3.0)
    serial_ccd = arctic_params.CCD(well_notch_depth=0.00, well_fill_alpha=1.0,
                                     well_fill_beta=0.5, well_fill_gamma=0.0)
    cti_params = arctic_params.ArcticParams(serial_ccd=serial_ccd, serial_species=[serial_species])

    serial_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=2, n_levels=2000,
                                                 charge_injection_mode=False, readout_offset=0)
    cti_settings = arctic_settings.ArcticSettings(serial=serial_settings)

    for data_resolution in data_resolutions:

        shape = simulation_util.shape_from_ci_data_resolution(ci_data_resolution=data_resolution)

        for normalization in normalizations:

            ci_regions = simulation_util.ci_regions_from_ci_data_resolution(ci_data_resolution=data_resolution)
            pattern = ci_pattern.CIPatternUniform(normalization=normalization, regions=ci_regions)

            cosmic_ray_image = cosmic_ray_image_from_shape_and_well_depth(shape=shape,
                                                                          well_depth=cti_settings.serial.well_depth)

            simulate_ci_data_from_ci_normalization_region_and_cti_model(
                ci_data_type=ci_data_type, ci_data_model=ci_data_model, ci_data_resolution=data_resolution,
                pattern=pattern, cti_params=cti_params, cti_settings=cti_settings, cosmic_ray_image=cosmic_ray_image)

def make_ci_uniform_cosmic_rays_parallel_and_serial_x1_species(data_resolutions, normalizations):

    ci_data_type = 'ci_uniform_cosmic_rays'
    ci_data_model = 'parallel_and_serial_x1'

    parallel_species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.5)
    parallel_ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0,
                                     well_fill_beta=0.5, well_fill_gamma=0.0)

    serial_species = arctic_params.Species(trap_density=1.0, trap_lifetime=3.0)
    serial_ccd = arctic_params.CCD(well_notch_depth=0.00, well_fill_alpha=1.0,
                                     well_fill_beta=0.5, well_fill_gamma=0.0)

    cti_params = arctic_params.ArcticParams(parallel_ccd=parallel_ccd, parallel_species=[parallel_species],
                                            serial_ccd=serial_ccd, serial_species=[serial_species])

    parallel_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=2, n_levels=2000,
                                                 charge_injection_mode=False, readout_offset=0)

    serial_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=2, n_levels=2000,
                                                 charge_injection_mode=False, readout_offset=0)

    cti_settings = arctic_settings.ArcticSettings(parallel=parallel_settings, serial=serial_settings)

    for data_resolution in data_resolutions:

        shape = simulation_util.shape_from_ci_data_resolution(ci_data_resolution=data_resolution)

        for normalization in normalizations:

            ci_regions = simulation_util.ci_regions_from_ci_data_resolution(ci_data_resolution=data_resolution)
            pattern = ci_pattern.CIPatternUniform(normalization=normalization, regions=ci_regions)

            cosmic_ray_image = cosmic_ray_image_from_shape_and_well_depth(shape=shape,
                                                                          well_depth=cti_settings.serial.well_depth)

            simulate_ci_data_from_ci_normalization_region_and_cti_model(
                ci_data_type=ci_data_type, ci_data_model=ci_data_model, ci_data_resolution=data_resolution,
                pattern=pattern, cti_params=cti_params, cti_settings=cti_settings, cosmic_ray_image=cosmic_ray_image)

def cosmic_ray_image_from_shape_and_well_depth(shape, well_depth):
    # We use the LA Cosmic algorithm to simulate and add cosmic rays to our ci pre cti image. This routine randomly
    # generates cosmimc rays based on realistic cosmic ray rates expected. These cosmic rays will then be added to our
    # ci pre-cti image in the simulaate function below, and subject to CTI according to the CTI model.
    cosmic_path = path_util.make_and_return_path_from_path_and_folder_names(path=workspace_path,
                                                                            folder_names=['scripts',
                                                                                          'cosmic_rays'])

    cosmic_ray_maker = cosmics.CosmicRays(shape=shape, cr_fluxscaling=1.0,
                                          cr_length_file=cosmic_path + 'crlength_v2.fits',
                                          cr_distance_file=cosmic_path + 'crdist.fits', log=logger)
    cosmic_ray_maker.set_ifiles()
    return cosmic_ray_maker.drawEventsToCoveringFactor(limit=well_depth)